# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: linetrace=False
# distutils: define_macros=CYTHON_TRACE_=0


from libcpp.algorithm cimport sort
from libc.math cimport isnan
from libcpp cimport bool as bool_t
from libcpp.map cimport map as cmap
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.set  cimport set
from cython.operator import dereference
import numpy as np
cimport numpy as c_np
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.dataaccess.services.transaction_file_record_reader import TransactionFileRecordReader

cdef:
    int DENSITY_INDEX = 0
    int MIN_INDEX = 1
    int MAX_INDEX = 2
    int SUM_INDEX = 3
    int FREQUENCY_INDEX = 4
    int COUNT_DISTINCT_INDEX = 5
    int MODE_INDEX = 6
    int TIME_SINCE_FIRST_INDEX = 7
    int TIME_SINCE_LAST_INDEX = 8

    c_np.float32_t NAN_VAL = np.nan
    agg_index_map = {
        AfeOperator.Density.value:0,
        AfeOperator.Min.value:1,
        AfeOperator.Max.value:2,
        AfeOperator.Sum.value:3,
        AfeOperator.Frequency.value:4,
        AfeOperator.CountDistinct.value:5,
        AfeOperator.Mode.value:6,
        AfeOperator.TimeSinceFirst.value:7,
        AfeOperator.TimeSinceLast.value:8,

    }
    cmap[int, int] AGGREGATE_INDEX_MAP = agg_index_map
    int RATIO_KEY = AfeOperator.Ratio.value
    int TIMESINCEFIRST_KEY = AfeOperator.TimeSinceFirst.value
    int TIMESINCELAST_KEY = AfeOperator.TimeSinceLast.value
    int SUM_KEY = AfeOperator.Sum.value
    int DENSITY_KEY = AfeOperator.Density.value
    int FREQUENCY_KEY = AfeOperator.Frequency.value
    int COUNT_DISTINCT_KEY = AfeOperator.CountDistinct.value
    int MODE_KEY = AfeOperator.Mode.value

    int UNKNOWN_DIMENSION_VALUE = TransactionFileRecordReader.UNKNOWN_DIMENSION_VALUE

def execute(
        float[:,:] result_frame,
        int c_num_threads,
        int date_col_index,
        int d_index,
        int q_index,
        int[:,:] splits,
        vector[int] horizons,
        int offset,
        int resolution,
        cmap[short, short] &index_map,
        c_np.int64_t[:,:] &trx_dates,
        c_np.float32_t[:,:] &trx_q_arrays,
        c_np.int8_t[:,:] &trx_d_arrays,
        cmap[int, c_np.int32_t[:]] &entity_index_map,
        c_np.int64_t[:] &target_dates ,
        vector[int] included_operators,
        vector[int] included_operator_set_one,
        vector[int] included_operator_set_two,
        cmap[long long, int] &column_name_map,
        int nearest_exponent,
        int dim_group_id = -1,
        bool_t for_afe_columns = False,
        bool_t is_no_dim_col = False
        ):
    cdef :
        int splits_size = splits.shape[0]
        int i

    for i in range(splits_size):
        exec_for_split(
            result_frame,
            i,
            splits_size,
            splits[i],
            date_col_index,
            d_index, q_index,
            offset,
            horizons,
            resolution,
            index_map,
            trx_dates,
            trx_q_arrays,
            trx_d_arrays,
            entity_index_map,
            target_dates,
            included_operators,
            included_operator_set_one,
            included_operator_set_two,
            column_name_map,
            nearest_exponent,
            dim_group_id,
            for_afe_columns,
            is_no_dim_col
            )

cdef void exec_for_split(
        float[:,:] &result_frame,
        int split_index,
        int split_count,
        int[:] &pair_list,
        int date_col_index,
        int d_index,
        int q_index,
        int offset,
        vector[int] &horizons,
        int resolution,
        cmap[short, short] &index_map,
        c_np.int64_t[:,:] &trx_dates,
        c_np.float32_t[:,:] &trx_q_arrays,
        c_np.int8_t[:,:] &trx_d_arrays,
        cmap[int, c_np.int32_t[:]] &entity_index_map,
        c_np.int64_t[:] &target_dates,
        vector[int] &included_operators,
        vector[int] &included_operator_set_one,
        vector[int] &included_operator_set_two,
        cmap[long long, int] &column_name_map,
        int nearest_exponent,
        int dim_group_id,
        bool_t is_scoring,
        bool_t is_no_dim_col
        ) nogil :

    cdef:
        int index, record_index, entity_id_int, pair_index
        c_np.int64_t event_date
        set[int] hor_set = set[int]()
        vector[int] local_horizon_list = horizons
        set[short] distinct_index_map_values = set[short]()
    get_distinct_values(distinct_index_map_values, index_map)
    cdef int s = distinct_index_map_values.size()

    cdef int h, th, distinct_horizon
    if is_scoring:
        for h in range(horizons.size()):
            hor_set.insert(horizons[h])
        local_horizon_list.clear()
        for distinct_horizon in hor_set:
            local_horizon_list.push_back(distinct_horizon)

    cdef:
        int horizon_count = local_horizon_list.size()
        float[:,:,:] _dict
    with gil:
        _dict = np.full((s, horizon_count, 9), NAN_VAL, dtype=np.float32)

    cdef:
        int min_horizon = get_min(local_horizon_list)
        int max_horizon = get_max(local_horizon_list)
        int max_horizon_for_fill_row = get_max(horizons)

    for pair_index in range(pair_list.shape[0]):
        index = split_index+pair_index*split_count

        record_index = pair_list[pair_index]

        if record_index == -1:
            break
        entity_id_int = record_index
        event_date = target_dates[record_index]

        if entity_index_map.count(entity_id_int) > 0:
                get_aggregates_and_fill_row(_dict,
                                    index, result_frame, entity_id_int, event_date, horizons,
                                    index_map, entity_index_map[entity_id_int], resolution, offset,
                                    trx_dates, trx_q_arrays, trx_d_arrays, date_col_index, d_index, q_index,
                                    included_operators,
                                    included_operator_set_one, included_operator_set_two,
                                    column_name_map, nearest_exponent, local_horizon_list,
                                    distinct_index_map_values, min_horizon, max_horizon, max_horizon_for_fill_row,
                                    dim_group_id, is_scoring, is_no_dim_col)



    #printf("%d %d %d\n", min_horizon, max_horizon, entity_index_map.size())

cdef void get_aggregates_and_fill_row(
        float[:,:,:] _dict,
        int row_index,
        float[:,:] &result_frame,
        int entity_id_int,
        c_np.int64_t event_date,
        vector[int] &horizons,
        cmap[short, short] &index_map,
        c_np.int32_t[:] &entity_index_list,
        int resolution, int offset,
        c_np.int64_t[:,:] &trx_dates,
        c_np.float32_t[:,:] &trx_q_arrays,
        c_np.int8_t[:,:] &trx_d_arrays,
        int date_col_index,
        int d_index,
        int q_index,
        vector[int] &included_operators,
        vector[int] &included_operator_set_one,
        vector[int] &included_operator_set_two,
        cmap[long long, int] &column_name_map,
        int nearest_exponent,
        vector[int] &local_horizon_list,
        set[short] &distinct_index_map_values,
        int min_horizon,
        int max_horizon,
        int max_horizon_for_fill_row,
        int dim_group_id,
        bool_t is_scoring,
        bool_t is_no_dim_col
        ) nogil :

    if entity_index_list.shape[0] == 0:
        return

    cdef:
        cmap[short, pair[float, float]] time_histogram
        cmap[int, double] mode_by_horizon
        cmap[int, double] count_distinct_by_horizon
        int curr_horizon = -1
        cmap[short, int] c_val_indices = cmap[short, int]()
        cmap[int, short] c_val_reverse_indices = cmap[int, short]()
        cmap[int, int] horizon_indices = cmap[int, int]()
        cmap[int, int] horizon_reverse_indices = cmap[int, int]()
        cmap[short, cmap[int, short]] calculated_dims_and_horizons = cmap[short, cmap[int, short]]()
        cdef set[int] hor_set = set[int]()

    _dict[:,:,:] =  NAN_VAL

    cdef:
        int s = distinct_index_map_values.size()
        int horizon_count = local_horizon_list.size()
        c_np.int64_t grid_last_date = event_date
        c_np.int64_t grid_first_date = event_date
        vector[int] grid_horizons = local_horizon_list
        int grid_max_horizon = max_horizon
        c_np.int64_t grid_offset_last_date = 0
        vector[pair[c_np.int64_t, int]] grid_dates_ascended = vector[pair[c_np.int64_t, int]]()

    partition(grid_dates_ascended, &grid_offset_last_date, &grid_first_date, grid_last_date, offset, resolution,
              grid_horizons, grid_max_horizon)

    cdef:
        int index = 0
        pair[short, short] c_val_pair
        short val
    for val in distinct_index_map_values:
        c_val_indices[val] = index
        c_val_reverse_indices[index] = val
        calculated_dims_and_horizons[val] = cmap[int, short]()
        index += 1

    #printf("Entity index list size for %d : %d\n", entity_id_int, entity_index_list.shape[0])

    cdef:
        int horizon_index = 0
        int tmp_hor
        int lhli
    for lhli in range(local_horizon_list.size()):
        tmp_hor = local_horizon_list[lhli]
        horizon_indices[tmp_hor] = horizon_index
        horizon_reverse_indices[horizon_index] = tmp_hor
        horizon_index += 1

    cdef:
        int record_index, d_val, c_val_index, horizon, hor_index, last_horizon, i, j, hor, entity_index_list_i
        c_np.int64_t record_date
        c_np.float32_t q_val
        short c_val, tmp_c_val
        float[:] pre
        float[:] *curr_dict
        float span
        bool_t ratio_op_exists = contains(included_operators, RATIO_KEY)
        bool_t count_distinct_op_exists = contains(included_operators, COUNT_DISTINCT_KEY)
        bool_t mode_op_exists = contains(included_operators, MODE_KEY)
        bool_t check_for_dimension = is_scoring and not ratio_op_exists and not is_no_dim_col and not (count_distinct_op_exists or mode_op_exists)
        cmap[int, int] d_val_dict = cmap[int, int]()
    # For RATIO operator, all dimensions should be aggregated. Dimension group should not be checked if dimension_column is "NOD"


    for entity_index_list_i in range(entity_index_list.shape[0]):

        record_index = entity_index_list[entity_index_list_i]
        record_date = 0

        if date_col_index != -1:
            record_date = trx_dates[record_index][date_col_index]
            if is_scoring:
                if record_date > grid_offset_last_date:
                    continue
            else:
                if record_date >= grid_offset_last_date:
                    continue

            if record_date < grid_first_date:
                break

        q_val = trx_q_arrays[record_index][q_index]

        if is_scoring:
            if isnan(q_val) or q_val <= 0:
                continue
        else:
            if q_val <= 0:
                continue

        d_val = trx_d_arrays[record_index][d_index]
        if count_distinct_op_exists or mode_op_exists:
            if d_val_dict.count(d_val) > 0:
                d_val_dict[d_val] = d_val_dict[d_val] + 1
            else:
                d_val_dict[d_val] = 1
        c_val = index_map[d_val]
        c_val_index = c_val_indices[c_val]
        #with gil:
            #print("%d %d %d\n", d_val, c_val, c_val_index)
        if  date_col_index != -1:
            horizon = get_horizon(grid_offset_last_date, grid_dates_ascended, record_date, is_scoring)
        else:
            horizon = 1

        if check_for_dimension and c_val != dim_group_id:
            continue

        #printf("Record : %d %d %d %f\n", entity_id_int, horizon, c_val, q_val)


        hor_index = horizon_indices[horizon]


        span = get_span(resolution, grid_last_date, record_date)

        if time_histogram.count(c_val) > 0:
            time_histogram[c_val] = pair[float,float](max(time_histogram[c_val].first, span), min(time_histogram[c_val].second, span))
        else:
            time_histogram[c_val] = pair[float, float](span, span)

        if horizon != curr_horizon:
            if curr_horizon != -1:
                for i in range(s):
                    tmp_c_val = c_val_reverse_indices[i]
                    if calculated_dims_and_horizons.count(tmp_c_val) > 0 and calculated_dims_and_horizons[tmp_c_val].size() > 0:
                        last_horizon = dereference(calculated_dims_and_horizons[tmp_c_val].rbegin()).first
                        pre = _dict[i][horizon_indices[last_horizon]]
                        for j in range(local_horizon_list.size()):  # pylint: disable=not-an-iterable
                            hor = local_horizon_list[j]
                            if horizon > hor > last_horizon:
                                _dict[i][horizon_indices[hor]] = pre
                        _dict[i][horizon_indices[horizon]] = pre

                last_horizon = dereference(count_distinct_by_horizon.rbegin()).first
                for j in range(local_horizon_list.size()):  # pylint: disable=not-an-iterable
                    hor = local_horizon_list[j]
                    if horizon > hor > last_horizon:
                        count_distinct_by_horizon[horizon_indices[hor]] =  count_distinct_by_horizon[last_horizon]
                        mode_by_horizon[horizon_indices[hor]] = mode_by_horizon[last_horizon]

            if calculated_dims_and_horizons[c_val].size() > 0:
                _dict[c_val_index][hor_index][MIN_INDEX] = min( _dict[c_val_index][hor_index][MIN_INDEX], q_val)
                _dict[c_val_index][hor_index][MAX_INDEX] = max( _dict[c_val_index][hor_index][MAX_INDEX], q_val)
                _dict[c_val_index][hor_index][SUM_INDEX] += q_val
                _dict[c_val_index][hor_index][FREQUENCY_INDEX] += 1
                _dict[c_val_index][hor_index][DENSITY_INDEX] =  _dict[c_val_index][hor_index][SUM_INDEX] /  _dict[c_val_index][hor_index][FREQUENCY_INDEX]

            else:
                _dict[c_val_index][hor_index][DENSITY_INDEX] = q_val
                _dict[c_val_index][hor_index][MIN_INDEX] = q_val
                _dict[c_val_index][hor_index][MAX_INDEX] = q_val
                _dict[c_val_index][hor_index][SUM_INDEX] = q_val
                _dict[c_val_index][hor_index][FREQUENCY_INDEX] = 1.0
            curr_horizon = horizon

        else:
            if calculated_dims_and_horizons[c_val].size() > 0:
                _dict[c_val_index][hor_index][MIN_INDEX] = min( _dict[c_val_index][hor_index][MIN_INDEX], q_val)
                _dict[c_val_index][hor_index][MAX_INDEX] = max( _dict[c_val_index][hor_index][MAX_INDEX], q_val)
                _dict[c_val_index][hor_index][SUM_INDEX] += q_val
                _dict[c_val_index][hor_index][FREQUENCY_INDEX] += 1
                _dict[c_val_index][hor_index][DENSITY_INDEX] =  _dict[c_val_index][hor_index][SUM_INDEX] /  _dict[c_val_index][hor_index][FREQUENCY_INDEX]
            else:
                _dict[c_val_index][hor_index][DENSITY_INDEX] = q_val
                _dict[c_val_index][hor_index][MIN_INDEX] = q_val
                _dict[c_val_index][hor_index][MAX_INDEX] = q_val
                _dict[c_val_index][hor_index][SUM_INDEX] = q_val
                _dict[c_val_index][hor_index][FREQUENCY_INDEX] = 1.0
        if count_distinct_op_exists :
            count_distinct_by_horizon[hor_index] = d_val_dict.size()
        if mode_op_exists:
            mode_by_horizon[hor_index] = get_mode(d_val_dict)

        calculated_dims_and_horizons[c_val][horizon] = True

    if curr_horizon < max_horizon:
        for i in range(s):
            tmp_c_val = c_val_reverse_indices[i]

            if calculated_dims_and_horizons.count(tmp_c_val)>0 and calculated_dims_and_horizons[tmp_c_val].size() > 0:
                last_horizon = dereference(calculated_dims_and_horizons[tmp_c_val].rbegin()).first
                pre = _dict[i][horizon_indices[last_horizon]]
                for j in range(local_horizon_list.size()):  # pylint: disable=not-an-iterable
                    hor = local_horizon_list[j]
                    if max_horizon >= hor > curr_horizon:
                        _dict[i][horizon_indices[hor]] = pre

        last_horizon = dereference(count_distinct_by_horizon.rbegin()).first
        for j in range(local_horizon_list.size()):  # pylint: disable=not-an-iterable
            hor = local_horizon_list[j]
            if max_horizon >= hor > curr_horizon:
                count_distinct_by_horizon[horizon_indices[hor]] =  count_distinct_by_horizon[last_horizon]
                mode_by_horizon[horizon_indices[hor]] = mode_by_horizon[last_horizon]
    cdef:
        float tsf, tsl
        float[:,:] aggregates
        int aggregates_size = _dict.shape[0]
        float[:,:] aggregate_by_horizon
        float[:] aggregate
        pair[float, float] time_pair
        pair[short, pair[float, float]] time_key_pair


    for time_key_pair in time_histogram:
        time_pair = time_key_pair.second
        aggregates = _dict[c_val_indices[time_key_pair.first]]
        for hor_index_pair in horizon_indices:
            aggregates[hor_index_pair.second][TIME_SINCE_FIRST_INDEX] = time_pair.first
            aggregates[hor_index_pair.second][TIME_SINCE_LAST_INDEX] = time_pair.second
        _dict[c_val_indices[time_key_pair.first]] = aggregates


    for c_val_index in range(aggregates_size):
        c_val = c_val_reverse_indices[c_val_index]
        aggregate_by_horizon = _dict[c_val_index]

        for hor_index in range(aggregate_by_horizon.shape[0]):
            aggregate = aggregate_by_horizon[hor_index]
            if mode_op_exists :
                aggregate[MODE_INDEX] = mode_by_horizon[hor_index]

            if count_distinct_op_exists :
                aggregate[COUNT_DISTINCT_INDEX] = count_distinct_by_horizon[hor_index]
            aggregate_by_horizon[hor_index] = aggregate

    if is_scoring:
        if not ratio_op_exists:
            fill_row(result_frame, date_col_index, d_index, q_index,
                    row_index, _dict, local_horizon_list,
                    max_horizon_for_fill_row, min_horizon, is_scoring, c_val_reverse_indices, horizon_reverse_indices,
                    horizon_indices,
                    included_operators, included_operator_set_one, included_operator_set_two,
                    column_name_map, nearest_exponent, dim_group_id, is_no_dim_col
                    )
        else:
            fill_row_for_ratio_column(result_frame,  date_col_index, d_index, q_index, row_index, _dict,
            local_horizon_list, max_horizon, min_horizon, is_scoring,
            c_val_reverse_indices, horizon_reverse_indices, horizon_indices, column_name_map,
            nearest_exponent, dim_group_id, horizons[0],
            is_no_dim_col)
    else:
        fill_row(result_frame,  date_col_index, d_index, q_index, row_index, _dict, local_horizon_list,
                max_horizon, min_horizon, is_scoring, c_val_reverse_indices, horizon_reverse_indices, horizon_indices,
                included_operators, included_operator_set_one, included_operator_set_two,
                column_name_map, nearest_exponent, dim_group_id,is_no_dim_col
                )

cdef void fill_row_for_ratio_column(
    float[:,:] &frame,
    int date_col_index,
    int d_index,
    int q_index,
    int row_index,
    float[:,:,:] &aggregates, vector[int] &horizons,
    int max_horizon, int min_horizon, bool_t is_scoring,
    cmap[int, short] &c_val_reverse_indices,
    cmap[int, int] &horizon_reverse_indices,
    cmap[int, int] &horizon_indices,
    cmap[long long, int] &column_name_map,
    int nearest_exponent,
    int ratio_dimension_group,
    int ratio_horizon,
    bool_t is_no_dim_col
    ) nogil :
    cdef:
        int i, hor
        cmap[int, float] totals = cmap[int, float]()

    for i in range(horizons.size()):
        hor = horizons[i]
        totals.insert(pair[int, float](hor, 0.0))

    cdef:
        cmap[int, cmap[int, float]] ratio_hist = cmap[int, cmap[int, float]]()
        int c_val
        int aggregates_size = aggregates.shape[0]
        float sum_val, i_value
        float[:,:] aggregate_by_horizon
        float[:] aggregate
        float total_val

    for c_val_index in range(aggregates_size):
        c_val = c_val_reverse_indices[c_val_index]
        aggregate_by_horizon = aggregates[c_val_index]
        ratio_hist[c_val] = cmap[int, float]()

        for hor_index in range(aggregate_by_horizon.shape[0]):
            horizon = horizon_reverse_indices[hor_index]
            aggregate = aggregate_by_horizon[hor_index]
            sum_val = aggregate[SUM_INDEX]
            if not isnan(sum_val):
                totals[horizon] += sum_val
            ratio_hist[c_val][horizon] = sum_val
    if ratio_hist.count(ratio_dimension_group) > 0 and ratio_hist[ratio_dimension_group].count(ratio_horizon) > 0:
        i_value = ratio_hist[ratio_dimension_group][ratio_horizon]
        total_val = totals[ratio_horizon]
        if not isnan(i_value) and total_val > 0 :
            frame_set_value(frame, column_name_map, row_index, i_value/total_val,  nearest_exponent,
            ratio_dimension_group, RATIO_KEY, date_col_index, ratio_horizon)


cdef void fill_row(
    float[:,:] &frame,
    int date_col_index,
    int  d_index,
    int q_index,
    int row_index,
    float[:,:,:] &aggregates, vector[int] &horizons,
    int max_horizon, int min_horizon, bool_t is_scoring,
    cmap[int, short] &c_val_reverse_indices,
    cmap[int, int] &horizon_reverse_indices,
    cmap[int, int] &horizon_indices,
    vector[int] &included_operators,
    vector[int] &included_operator_set_one,
    vector[int] &included_operator_set_two,
    cmap[long long, int] &column_name_map,
    int nearest_exponent,
    int dim_group_id,
    bool_t is_no_dim_col
    ) nogil :

    cdef:
        int min_horizon_index = horizon_indices[min_horizon]
        int i,  hor
        cmap[int, float] totals = cmap[int, float]()

    for i in range(horizons.size()):
        hor = horizons[i]
        totals.insert(pair[int, float](hor, 0.0))

    cdef:
        cmap[int, cmap[int, float]] ratio_hist = cmap[int, cmap[int, float]]()
        int c_val, horizon, column_index, op_name, time_window
        float[:,:] aggregate_by_horizon
        float[:] aggregate
        float sum_val, val, total_val
        int aggregates_size = aggregates.shape[0]
        bool_t count_distinct_op_exists = contains(included_operators, COUNT_DISTINCT_KEY)
        bool_t mode_op_exists = contains(included_operators, MODE_KEY)
        bool_t check_for_dimension = is_scoring and not is_no_dim_col and not ( count_distinct_op_exists or mode_op_exists)
        int op_i, c_val_index, hor_index

    for c_val_index in range(aggregates_size):
        c_val = c_val_reverse_indices[c_val_index]
        if check_for_dimension and c_val != dim_group_id:
            continue
        aggregate_by_horizon = aggregates[c_val_index]
        ratio_hist[c_val] = cmap[int, float]()

        for hor_index in range(aggregate_by_horizon.shape[0]):
            horizon = horizon_reverse_indices[hor_index]
            aggregate = aggregate_by_horizon[hor_index]
            sum_val = aggregate[SUM_INDEX]
            if not isnan(sum_val):
                totals[horizon] += sum_val
            ratio_hist[c_val][horizon] = sum_val
            for op_i in range(included_operator_set_one.size()):
                op_name = included_operator_set_one[op_i]
                val = aggregate[AGGREGATE_INDEX_MAP[op_name]]
                frame_set_value(frame,  column_name_map, row_index, val, nearest_exponent, c_val, op_name,
                                date_col_index, horizon)
        for op_i in range(included_operator_set_two.size()):
            op_name = included_operator_set_two[op_i]

            val = aggregate_by_horizon[min_horizon_index][AGGREGATE_INDEX_MAP[op_name]]
            frame_set_value(frame, column_name_map, row_index, val, nearest_exponent, c_val, op_name,
                            date_col_index, horizon)

    if not is_scoring:
        if contains(included_operators,RATIO_KEY):

            for c_val_val_pair in ratio_hist:
                for hor_val_pair in c_val_val_pair.second:
                    total_val = totals[hor_val_pair.first]
                    if not isnan(hor_val_pair.second) and total_val > 0 :
                        frame_set_value(frame,  column_name_map, row_index, hor_val_pair.second/total_val,
                        nearest_exponent, c_val_val_pair.first, RATIO_KEY, date_col_index, hor_val_pair.first)



cdef void frame_set_value(float[:,:] &frame, cmap[long long, int] &column_name_map, int row_index, float val,
            int nearest_exponent, int group_id, int afe_operator, int date_col_index, int horizon) nogil :
    if group_id == UNKNOWN_DIMENSION_VALUE:
        return
    cdef:
        int hash_val  = get_hash_value(nearest_exponent, group_id, afe_operator, horizon)
        int column_index = column_name_map[hash_val]
    frame[row_index][column_index] = val

cdef int get_hash_value(int nearest_exponent, int group_id, int afe_operator, int horizon) nogil :
    cdef:
        int val = nearest_exponent ** 2 * group_id + nearest_exponent * afe_operator + horizon
    return val

cdef bool_t contains(vector[int] &vec, int val) nogil :
    cdef:
        int i, v
    for i in range(vec.size()):
        v = vec[i]
        if val == v:
            return True
    return False

cdef set[short] get_distinct_values(set[short] &s, cmap[short, short] &_map) nogil :
    cdef:
        pair[short, short] p
    for p in _map:
        s.insert(p.second)
    return s

cdef c_np.int64_t add_offset(c_np.int64_t to_sub, c_np.int64_t last_date, int step) nogil :
    return last_date - to_sub*step

cdef c_np.int64_t get_resolution_in_seconds(int resolution) nogil :
    cdef c_np.int64_t to_sub = 1000 #overflow olmaması için int64 tanımlanıp onun üzerinden işlem yapılmalı
    if resolution == 0:
        to_sub *=  (365*24*60*60)
    elif resolution == 1:
        to_sub *= (30*24*60*60)
    elif resolution == 2:
        to_sub *= (24*60*60)
    elif resolution == 3:
        to_sub *= (60*60)
    elif resolution == 4:
        to_sub *= (60)
    elif resolution == 5:
        to_sub *= 1
    else:
        raise NotImplementedError("Resolution not recognized: " + resolution)
    return to_sub

cdef float get_span(int resolution, c_np.int64_t last_date,  c_np.int64_t dt) nogil :
    cdef:
        c_np.int64_t val = last_date-dt
        c_np.int64_t to_divide = 1000
    if resolution == 0:
        to_divide *= (365*24*60*60)
    elif resolution == 1:
        to_divide *= (30*24*60*60)
    elif resolution == 2:
        to_divide *= (24*60*60)
    elif resolution == 3:
        to_divide *= (60*60)
    elif resolution == 4:
        to_divide *= (60)
    elif resolution == 5:
        to_divide *= 1
    else:
        raise NotImplementedError("Resolution not recognized: " + resolution)
    return val*1.0 / to_divide

cdef void partition(vector[pair[c_np.int64_t, int]] &dates_ascended, c_np.int64_t *offset_last_date,
                    c_np.int64_t *first_date, c_np.int64_t last_date, int offset, int resolution,
                    vector[int] &horizons, int max_horizon) nogil :
    cdef:
        c_np.int64_t resolution_in_seconds = get_resolution_in_seconds(resolution)
        int i, j
    offset_last_date[0] = add_offset(resolution_in_seconds , last_date, offset)
    dates_ascended.push_back(pair[c_np.int64_t, int](offset_last_date[0], horizons[0]))
    cdef int initial = horizons.size() - 1
    for j in range(initial):
        i = initial - j
        dates_ascended.push_back(pair[c_np.int64_t, int](add_offset(resolution_in_seconds, last_date, horizons[i - 1] + offset), horizons[i]))
    sort(dates_ascended.begin(), dates_ascended.end())
    first_date[0] = add_offset(resolution_in_seconds, last_date, max_horizon + offset)

cdef int get_horizon(c_np.int64_t offset_last_date, vector[pair[c_np.int64_t, int]] &dates_ascended,
    c_np.int64_t date, bool_t allow_equal_to_last_date=False) nogil :

    if allow_equal_to_last_date and date == offset_last_date:
        return dates_ascended[dates_ascended.size()-1].second
    for pair in dates_ascended:
        if date < pair.first:
            return pair.second
    raise ValueError("Date {0} is out of range".format(date))

cdef int get_min(vector[int] vec) nogil :
    cdef:
        int min_val = vec[0]
        int val
    for val in vec:
        if val < min_val:
            min_val = val
    return min_val

cdef int get_max(vector[int] vec) nogil :
    cdef:
        int val
        int max_val = vec[0]
    for val in vec:
        if val > max_val:
            max_val = val
    return max_val

cdef pair[int, int] find_entry_with_largest_value(cmap[int,int] map) nogil :
    cdef:
        pair[int, int] entryWithMaxValue
    entryWithMaxValue.first = 0
    entryWithMaxValue.second = 0

    for currentEntry in map:
            if (currentEntry.second> entryWithMaxValue.second) :
                entryWithMaxValue.first = currentEntry.first
                entryWithMaxValue.second = currentEntry.second

    return entryWithMaxValue


cdef int get_mode(cmap[int,int] map) nogil :
    cdef:
        pair[int, int] entry_with_max_value = find_entry_with_largest_value(map);
        int max_key = entry_with_max_value.first
    return max_key




"""
This class keeps the EnvironmentInfoDto class.
"""


class EnvironmentInfoDto:
    """
    This class keeps the environment info.
    """
    def __init__(self, machine_name: str, user_name: str):
        self.machine_name = machine_name
        self.user_name = user_name



"""Includes TransferLearningSettings class"""


class OptimizerSettings:
    """Settings for transfer learning in text classification"""

    def __init__(self, learning_rate: float = 1e-5, early_stopping: int = 5, steps_per_epoch: int = None,
                 early_stopping_min_delta: float = 1e-3):
        self.learning_rate: float = learning_rate
        self.early_stopping: int = early_stopping
        self.steps_per_epoch: int = steps_per_epoch
        self.early_stopping_min_delta: float = early_stopping_min_delta



"""Includes DqFileComparisonParameters class."""
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings


class DqFileComparisonParameters(DqBaseComparisonParameters[DqFileInputSourceSettings]):
    """DQ comparison parameters for file source type"""



"""
This module includes base class for afe report helpers.
"""
import abc

from organon.afe.domain.reporting.abstract_report_helper import AbstractReportHelper
from organon.afe.domain.reporting.afe_model_output import AfeModelOutput
from organon.afe.domain.reporting.afe_output_report import AfeOutputReport


class BaseReportHelper(AbstractReportHelper[AfeModelOutput, AfeOutputReport], metaclass=abc.ABCMeta):
    """
    This class is base abstract class for afe reporting.
    """

    @classmethod
    def generate_afe_report(cls, model_output: AfeModelOutput) -> AfeOutputReport:
        """Creates output report from model output and runtime statistics"""
        report = AfeOutputReport()
        report.model_identifier = model_output.model_identifier
        output_features = model_output.output_features
        report.output_features = output_features
        report.all_features = model_output.all_features
        report.dimension_names_map = cls._get_dimension_names_map(model_output)
        report.transformations = model_output.transformations
        report.execution_time = model_output.runtime_stats.execution_time
        report.memory_usage = model_output.runtime_stats.memory_usage
        report.cpu_usage = model_output.runtime_stats.cpu_usage
        report.event_times = model_output.runtime_stats.event_times
        report.feature_counts_report = cls._get_feature_counts_report(output_features)
        report.final_column_metrics = model_output.final_column_metrics
        return report



"""Includes BaseMultipleSourceReportHelper class."""
import abc
from typing import List, TypeVar, Generic

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.reporting.base_dq_report_helper import BaseDqReportHelper
from organon.idq.domain.reporting.objects.base_alert_info import BaseAlertInfo
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport
from organon.idq.domain.reporting.objects.base_multiple_calculation_report_helper_params import \
    BaseMultipleCalculationReportHelperParams
from organon.idq.domain.reporting.objects.dq_comparison_sample_report_input import DqComparisonSampleReportInput
from organon.idq.domain.reporting.objects.nominal_column_control_details import NominalColumnControlDetails
from organon.idq.domain.reporting.objects.numeric_column_control_details import NumericColumnControlDetails
from organon.idq.domain.reporting.objects.table_control_details import TableControlDetails

DqComparisonSampleReportInputType = TypeVar("DqComparisonSampleReportInputType", bound=DqComparisonSampleReportInput)


class BaseMultipleCalculationReportHelper(Generic[DqComparisonSampleReportInputType], BaseDqReportHelper):
    """Includes helper methods for BaseMultipleSourceReportHelper"""

    def __init__(self, base_multiple_source_dq_report_helper_params: BaseMultipleCalculationReportHelperParams):
        super().__init__(base_multiple_source_dq_report_helper_params)
        self.calculation_count = base_multiple_source_dq_report_helper_params.calculation_count
        self.comparison_column_info = base_multiple_source_dq_report_helper_params.comparison_column_info
        self.sample_statistics = base_multiple_source_dq_report_helper_params.sample_statistics
        self.population_statistics = base_multiple_source_dq_report_helper_params.population_statistics
        self.data_source_statistics = base_multiple_source_dq_report_helper_params.data_source_statistics

    @abc.abstractmethod
    def execute(self) -> BaseDqOutputReport:
        """Execute reporting for one df"""
        raise NotImplementedError

    @abc.abstractmethod
    def _get_table_alert(self, table_comparison_result):
        """Return alert of table"""
        raise NotImplementedError

    @staticmethod
    def _get_numeric_alert_list(calculation_count: int) -> List[DqComparisonResultCode]:
        if calculation_count >= 5:
            numeric_alert_list = [DqComparisonResultCode.COLUMN_MEAN_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM,
                                  DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL,
                                  DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL,
                                  DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL,
                                  DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL,
                                  DqComparisonResultCode.PSI_RED_SIGNAL
                                  ]
        else:
            numeric_alert_list = [DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL,
                                  DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL,
                                  DqComparisonResultCode.PSI_RED_SIGNAL
                                  ]
        return numeric_alert_list

    @staticmethod
    def _get_column_stability_alert_list() -> List[DqComparisonResultCode]:
        alert_list = [DqComparisonResultCode.COLUMN_IS_CONSTANT,
                      DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_ZERO,
                      DqComparisonResultCode.COLUMN_VALUES_ARE_ALL_NULL]
        return alert_list

    @staticmethod
    def _get_table_alert_list(calculation_count: int) -> List[DqComparisonResultCode]:
        if calculation_count >= 5:
            table_alert_list = [DqComparisonResultCode.ROW_COUNT_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM,
                                DqComparisonResultCode.ROW_COUNT_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL,
                                DqComparisonResultCode.ROW_COUNT_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL,
                                DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL,
                                DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL
                                ]
        else:
            table_alert_list = [DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL,
                                DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL]
        return table_alert_list

    @abc.abstractmethod
    def _get_numeric_column_alert(self, numeric_comparison_result):
        """Return numeric column alert of table"""
        raise NotImplementedError

    @abc.abstractmethod
    def _get_nominal_column_alert(self, nominal_comparison_result):
        """Return nominal column alert of table"""
        raise NotImplementedError

    def _get_alert_data(self) -> List[BaseAlertInfo]:
        """Return alert data multiple data sources"""
        list_base_alert_info: List[BaseAlertInfo] = []
        table_comparison_result, numeric_comparison_result, nominal_comparison_result = self._get_result_with_type()
        table_comparison_result_list = [result.result_code for result in table_comparison_result]
        table_alert = self._get_table_alert(table_comparison_result_list)
        list_base_alert_info.extend(table_alert)
        numeric_column_alert = self._get_numeric_column_alert(numeric_comparison_result)
        list_base_alert_info.extend(numeric_column_alert)
        nominal_column_alert = self._get_nominal_column_alert(nominal_comparison_result)
        list_base_alert_info.extend(nominal_column_alert)
        return list_base_alert_info

    def _get_report_data(self) -> DqComparisonSampleReportInputType:

        report_data = self._get_report_input_instance()
        report_data.dq_data_column_collection = self.data_column_collection
        report_data.dq_comparison_result = self.comparison_results
        alert_info_list = self._get_alert_data()
        report_data.alert_info = alert_info_list
        report_data.numeric_columns_control = self._get_numeric_column_details(alert_info_list)
        report_data.nominal_columns_control = self._get_nominal_column_details(alert_info_list)
        report_data.table_control = self._get_table_control_details(alert_info_list)

        return report_data

    @classmethod
    def _get_report_input_instance(cls) -> DqComparisonSampleReportInputType:
        return DqComparisonSampleReportInput()

    def _get_numeric_column_details(self, alert_info: List[BaseAlertInfo]) \
            -> List[NumericColumnControlDetails]:
        """Return numeric column details"""
        list_numeric_column_details: List[NumericColumnControlDetails] = []
        alert_column_list = {alert.alert for alert in alert_info}
        for state_of_benchmark, (calculation_name, sample_stats) in enumerate(self.sample_statistics.items()):
            numeric_stats_dict = sample_stats.numerical_statistics
            for column_name, numeric_stats in numeric_stats_dict.items():
                if column_name in alert_column_list:
                    numeric_column_details = NumericColumnControlDetails()
                    numeric_column_details.calculation_name = calculation_name
                    numeric_column_details.column_name = column_name
                    numeric_column_details.column_type = ColumnNativeType.Numeric
                    numeric_column_details.mean = numeric_stats.mean
                    numeric_column_details.trimmed_mean = numeric_stats.trimmed_mean
                    numeric_column_details.state_of_benchmark = state_of_benchmark + 1
                    list_numeric_column_details.append(numeric_column_details)
        return list_numeric_column_details

    def _get_nominal_column_details(self, alert_info: List[BaseAlertInfo]) \
            -> List[NominalColumnControlDetails]:
        """Return nominal column details"""
        list_nominal_column_details: List[NominalColumnControlDetails] = []
        alert_column_list = {alert.alert for alert in alert_info}

        for state_of_benchmark, (calculation_name, population_stats) in enumerate(self.population_statistics.items()):
            population_statistics = population_stats.nominal_statistics
            for column_name, nominal_stats in population_statistics.items():
                if column_name in alert_column_list:
                    data_column = self.data_column_collection.get_column(column_name)
                    value = self._get_nominal_value_for_column(column_name)
                    nominal_column_details = NominalColumnControlDetails()
                    nominal_column_details.calculation_name = calculation_name
                    nominal_column_details.column_name = column_name
                    nominal_column_details.column_type = data_column.column_native_type
                    if value in nominal_stats.percentages:
                        nominal_column_details.percentage = nominal_stats.percentages[value]
                    else:
                        nominal_column_details.percentage = 0
                    nominal_column_details.state_of_benchmark = state_of_benchmark + 1
                    nominal_column_details.value = value
                    list_nominal_column_details.append(nominal_column_details)
        return list_nominal_column_details

    def _get_table_control_details(self, alert_info: List[BaseAlertInfo]) -> List[TableControlDetails]:
        """Return table control details"""
        list_table_control_details: List[TableControlDetails] = []
        alert_column_list = {alert.alert for alert in alert_info}
        if DataEntityType.TABLE.name in alert_column_list:
            for state_of_benchmark, (calculation_name, table_stats) in enumerate(self.data_source_statistics.items()):
                table_size = table_stats.size_in_bytes
                row_count = table_stats.row_count
                if table_size is not None:
                    if table_size > 0:
                        table_control_details = TableControlDetails()
                        table_control_details.calculation_name = calculation_name
                        table_control_details.type = DataEntityType.TABLE
                        table_control_details.property_key = "TableSizeInBytes"
                        table_control_details.value = table_size
                        table_control_details.state_of_benchmark = state_of_benchmark + 1
                        list_table_control_details.append(table_control_details)
                if row_count is not None:
                    if row_count > 0:
                        table_control_details = TableControlDetails()
                        table_control_details.calculation_name = calculation_name
                        table_control_details.type = DataEntityType.TABLE
                        table_control_details.property_key = "RowCount"
                        table_control_details.value = row_count
                        table_control_details.state_of_benchmark = state_of_benchmark + 1
                        list_table_control_details.append(table_control_details)
        return list_table_control_details

    def _get_nominal_value_for_column(self, column_name: str) -> List[str]:
        """Return value  of nominal columns"""
        value = None
        for comp in self.comparison_results:
            if comp.data_entity_name == column_name and comp.property_code == "MaximumPsiCategory":
                value = comp.property_key_nominal
        return value

    @staticmethod
    def _get_alert_df(report_data: DqComparisonSampleReportInputType):
        alert_list = report_data.alert_info
        alerts_df = pd.DataFrame.from_records([a.to_dict() for a in alert_list])
        alerts_df = alerts_df.drop_duplicates(
            subset=alerts_df.columns,
            keep='last').reset_index(drop=True)
        return alerts_df

    @staticmethod
    def _get_signal_df(report_data: DqComparisonSampleReportInputType):
        signal_list = BaseMultipleCalculationReportHelper._get_signal_details(report_data)
        signals_df = pd.DataFrame.from_records([s.to_dict() for s in signal_list])
        signals_df = signals_df.drop_duplicates(
            subset=signals_df.columns,
            keep='last').reset_index(drop=True)
        return signals_df

    @staticmethod
    def _get_numeric_columns_control_df(report_data: DqComparisonSampleReportInputType):
        numeric_columns_control_list = report_data.numeric_columns_control
        numeric_columns_control_df = pd.DataFrame.from_records([num.to_dict() for num in numeric_columns_control_list])
        return numeric_columns_control_df

    @staticmethod
    def _get_nominal_columns_control_df(report_data: DqComparisonSampleReportInputType):
        nominal_columns_control_list = report_data.nominal_columns_control
        nominal_columns_control_df = pd.DataFrame.from_records([nom.to_dict() for nom in nominal_columns_control_list])
        return nominal_columns_control_df

    @staticmethod
    def _get_table_control_df(report_data: DqComparisonSampleReportInputType):
        table_control_list = report_data.table_control
        table_control_df = pd.DataFrame.from_records([t.to_dict() for t in table_control_list])
        return table_control_df

    def _fill_output_report(self, report_data: DqComparisonSampleReportInputType, output_report):
        output_report.alerts_df = self._get_alert_df(report_data)
        output_report.signals_df = self._get_signal_df(report_data)
        output_report.signals = report_data.dq_comparison_result
        output_report.column_collection = report_data.dq_data_column_collection
        output_report.numeric_column_control_details_df = self._get_numeric_columns_control_df(report_data)
        output_report.nominal_column_control_details_df = self._get_nominal_columns_control_df(report_data)
        output_report.table_control_details_df = self._get_table_control_df(report_data)
        return output_report



"""Includes helper functions for afe date columns."""
from datetime import timedelta, datetime
from typing import Callable

from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.fl.core.helpers.date_helper import get_date_from_string, format_date


def get_date_subtraction_func(resolution: DateResolution, delta_value: int) -> Callable[[int], int]:
    """
    Returns a function to manipulate a date based on given resolution
    :param resolution:
    :type resolution: DateResolution
    :param delta_value: number of years,month,dates.. to add as delta
    :type delta_value: int
    :return: Function to manipulate date
    """
    step = -1 * delta_value

    function_map = {
        DateResolution.Year: timedelta(days=365 * step).total_seconds(),
        DateResolution.Month: timedelta(days=30 * step).total_seconds(),
        DateResolution.Day: timedelta(days=step).total_seconds(),  # timedelta is faster
        DateResolution.Hour: timedelta(hours=step).total_seconds(),
        DateResolution.Minute: timedelta(minutes=step).total_seconds(),
        DateResolution.Second: timedelta(seconds=step).total_seconds()
    }
    if resolution in function_map:
        delta = int(function_map[resolution] * 1000)
        return lambda timestamp: timestamp + delta

    raise NotImplementedError


def get_str_to_date_converter(date_column_type: AfeDateColumnType, date_custom_format: str = None) -> \
        Callable[[str], datetime]:
    """
    Returns a function that performs string to date conversion according to date column type and if necessary
    date custom format
    """
    if date_column_type == AfeDateColumnType.DateTime:
        return lambda s: get_date_from_string(s, iso_format=True)
    if date_column_type == AfeDateColumnType.YyyyMmDd:
        return lambda s: datetime(int(s[0:4]), int(s[4:6]), int(s[6:8]))
    if date_column_type == AfeDateColumnType.YyyyMm:
        return lambda s: datetime(int(s[0:4]), int(s[4:6]), 1)
    if date_column_type == AfeDateColumnType.CustomFormat:
        return lambda s: get_date_from_string(s, date_custom_format)
    raise NotImplementedError


def get_date_to_str_converter(date_column_type: AfeDateColumnType, date_custom_format: str = None) -> Callable[
    [datetime], str]:
    """Returns a function that will convert a given datetime object to string according to date_column_type"""
    if date_column_type == AfeDateColumnType.DateTime:
        return lambda s: format_date(s)  # pylint: disable=unnecessary-lambda
    if date_column_type == AfeDateColumnType.YyyyMmDd:
        return lambda s: format_date(s, "%Y%m%d")
    if date_column_type == AfeDateColumnType.YyyyMm:
        return lambda s: format_date(s, "%Y%m")
    if date_column_type == AfeDateColumnType.CustomFormat:
        return lambda s: format_date(s, date_custom_format)
    raise NotImplementedError



"""
This module keeps the get_current_directory function.
"""
import os
import shutil
from typing import List


def get_current_directory() -> str:
    """
    Returns the current working directory path.
    :return: current working directory path
    """
    path = os.getcwd()
    return path


def get_name(path: str) -> str:
    """
    Returns the name of the file in the given path
    :return: name of the file in the given path
    """
    return os.path.basename(path)


def exists(path: str) -> bool:
    """
    Returns True if given path exists
    """
    return os.path.isdir(path)


def create(path: str, exception_if_exists: bool = False):
    """
    Creates directories and subdirectories for given path
    """
    os.makedirs(path, exist_ok=not exception_if_exists)


def copy_files(file_paths: List[str], target_dir: str):
    """Copies given files to target directory"""
    create(target_dir, exception_if_exists=False)
    for file in file_paths:
        shutil.copy(file, os.path.join(target_dir, os.path.basename(file)))


def copy_dir(source_dir: str, target_dir: str):
    """Copies given directory to target directory"""
    shutil.copytree(source_dir, target_dir)


def delete_directory_with_contents(directory: str, ignore_errors: bool = False):
    """Deletes directory and all its contents"""
    shutil.rmtree(directory, ignore_errors)



"""Includes LockedIterator class"""
import threading


class LockedIterator:
    """Iterator which is locked on iteration (making it thread-safe)"""
    def __init__(self, iterator):
        self.lock = threading.Lock()
        self.iterator = iterator.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.iterator)



""" This module includes ColumnMeanControl"""
from typing import List

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.algorithms.objects.mean_ci_comparison_input import MeanConfidenceIntervalComparisonInput
from organon.idq.domain.algorithms.objects.time_series_comparison_input import TimeSeriesComparisonInput
from organon.idq.domain.algorithms.objects.traffic_light_comparison_input import TrafficLightComparisonInput
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.controls.helpers.traffic_light_comparer import TrafficLightComparer
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters


class ColumnMeanControl(BaseDqControl):
    """ ColumnMeanControl class"""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        """ :returns test group type of control"""
        return DqTestGroupType.MODELLING_AND_CI_CONTROLS_COLUMN_SET

    @staticmethod
    def get_control_type() -> DqControlType:
        """ :returns control type of control"""
        return DqControlType.COLUMN_MEAN

    def _execute_control(self) -> List[DqComparisonResult]:
        """ :returns list of comparison results"""
        results = []
        if self.comp_context.control_results is None:
            raise ValueError("Control results None")
        if self.comp_context.test_calc_result is None:
            raise ValueError("Test result None")
        numerical_comp_cols = self.comp_context.get_comparison_cols_by_native_type(ColumnNativeType.Numeric)
        test_group_info = self.comp_context.test_group_info[self.get_test_group_type()]
        for col in numerical_comp_cols:
            col_name = col.column_name
            effective_bmh = min(test_group_info.test_bmh, col.benchmark_horizon)
            max_index = max(len(self.comp_context.control_results) - effective_bmh, 0)
            filtered_controls = list(self.comp_context.control_results[max_index:])
            test_stat_collection = self.comp_context.test_calc_result.sample_stats.numerical_statistics
            if len(filtered_controls) > 0 and col_name in test_stat_collection:
                control_trimmed_means = [control.sample_stats.numerical_statistics[col_name].trimmed_mean for control in
                                         filtered_controls if col_name in control.sample_stats.numerical_statistics]
                if len(control_trimmed_means) == 0:
                    LogHelper.info(f"Numeric Controls skipped, no past value for {col_name} found")
                else:
                    test_numerical_stats = test_stat_collection[col_name]
                    test_mean = test_numerical_stats.trimmed_mean
                    mi_input = self.__get_mi_input(col_name, control_trimmed_means,
                                                   self.comp_context.comparison_parameters, test_mean)
                    tl_input = self.__get_tl_input(col_name, control_trimmed_means,
                                                   self.comp_context.comparison_parameters, test_mean)
                    ts_input = self.__get_ts_input(col_name, control_trimmed_means,
                                                   self.comp_context.comparison_parameters, test_mean, effective_bmh)
                    results.extend(TrafficLightComparer.get_mi_tl_comparison_results(mi_input, tl_input))
                    results.extend(TrafficLightComparer.get_dq_comparison_results_for_time_series(ts_input))
        return results

    def get_description(self) -> str:
        """ :returns description of control"""
        return "Column Mean Controls"

    @staticmethod
    def __get_mi_input(col: str, past_series: List[float],
                       comp_parameters: DqBaseComparisonParameters,
                       current_val: float) -> MeanConfidenceIntervalComparisonInput:
        """:returns mean confidence interval comparison input"""
        mi_input = MeanConfidenceIntervalComparisonInput()
        mi_input.data_entity = DataEntityType.COLUMN
        mi_input.data_entity_name = col
        mi_input.test_group = ColumnMeanControl.get_test_group_type()
        mi_input.past_series = past_series
        mi_input.current_value = current_val
        mi_input.z_score = comp_parameters.z_score
        mi_input.result_code = DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL
        mi_input.message = f"Column {col} mean is outside mean confidence interval."
        return mi_input

    @staticmethod
    def __get_tl_input(col: str, past_series: List[float],
                       comp_parameters: DqBaseComparisonParameters, current_val: float) -> TrafficLightComparisonInput:
        """ :returns traffic light comparer comparison input"""
        tl_input = TrafficLightComparisonInput()
        tl_input.data_entity = DataEntityType.COLUMN
        tl_input.data_entity_name = col
        tl_input.test_group = ColumnMeanControl.get_test_group_type()
        tl_input.past_series = past_series
        tl_input.current_value = current_val
        tl_input.green_threshold = comp_parameters.traffic_light_threshold_green
        tl_input.yellow_threshold = comp_parameters.traffic_light_threshold_yellow
        tl_input.result_code = DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL
        tl_input.message = f"Column {col} mean is outside mean traffic light interval."
        return tl_input

    @staticmethod
    def __get_ts_input(col: str, past_series: List[float], comp_parameters: DqBaseComparisonParameters,
                       current_val: float, bmh: int) -> TimeSeriesComparisonInput:
        """ :returns time series comparison input"""
        ts_input = TimeSeriesComparisonInput()
        ts_input.data_entity = DataEntityType.COLUMN
        ts_input.data_entity_name = col
        ts_input.test_group = ColumnMeanControl.get_test_group_type()
        ts_input.past_series = past_series
        ts_input.current_value = current_val
        ts_input.bmh = bmh
        ts_input.z_score = comp_parameters.z_score
        ts_input.green_threshold = comp_parameters.traffic_light_threshold_green
        ts_input.yellow_threshold = comp_parameters.traffic_light_threshold_yellow
        ts_input.ad_code = DqComparisonResultCode.COLUMN_MEAN_ABSOLUTE_DEVIATION_FROM_PREDICTION_IS_MAXIMUM
        ts_input.ci_code = DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_PREDICTION_CONFIDENCE_INTERVAL
        ts_input.tl_code = DqComparisonResultCode.COLUMN_MEAN_IS_OUTSIDE_PREDICTION_TRAFFIC_LIGHT_INTERVAL
        ts_input.ad_message = f"Column {col} absolute deviation from prediction is maximum."
        ts_input.ci_message = f"Column {col} is outside prediction confidence interval."
        ts_input.tl_message = f"Column {col} is outside prediction traffic light interval."
        return ts_input



"""settings class for word tokenization"""


class TokenizerSettings:
    """word tokenization parameters class"""

    def __init__(self, padding: str = "max_length", truncation: bool = True, max_length: int = 128):
        self.padding: str = padding
        self.truncation: bool = truncation
        self.max_length: int = max_length



"""
This module keeps the IocRegistrationItem class.
"""


class IocRegistrationItem:
    """
    Holds the features such as ioc key, abstract type and concrete type to be used in IoC
    """
    def __init__(self, ioc_key: str, abstract_type: type, concrete_type: type):
        self.ioc_key = ioc_key
        self.abstract_type = abstract_type
        self.concrete_type = concrete_type



"""Includes DqBaseInputSourceSettings class."""
import sys
from typing import TypeVar, Generic, List, Callable

import pandas as pd

from organon.idq.domain.businessobjects.record_sources.dq_base_record_source import DqBaseRecordSource
from organon.idq.domain.settings.partition_info import PartitionInfo

DqRecordSourceType = TypeVar("DqRecordSourceType", bound=DqBaseRecordSource)


class DqBaseInputSourceSettings(Generic[DqRecordSourceType]):
    """Base input source settings for input source types"""

    def __init__(self, source: DqRecordSourceType):
        self.source: DqRecordSourceType = source
        self.is_sampling_on: bool = None
        self.sampling_ratio: float = None
        self.max_num_of_samples: int = sys.maxsize
        self.included_columns: List[str] = None
        self.partition_info_list: List[PartitionInfo] = None
        self.filter_callable: Callable[[pd.DataFrame], pd.DataFrame] = None



"""Includes DqFullProcessInput class."""
from typing import List, TypeVar, Generic

from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters
from organon.idq.domain.settings.dask_settings import DaskSettings
from organon.idq.domain.settings.dq_notification_settings import DqNotificationSettings
from organon.idq.domain.settings.dq_output_settings import DqOutputSettings

T1 = TypeVar("T1", bound=DqBaseCalculationParameters)
T2 = TypeVar("T2", bound=DqBaseComparisonParameters)


class DqFullProcessInput(Generic[T1, T2]):
    """All settings required for a full DQ process."""

    def __init__(self):
        self.calculation_parameters: List[T1] = None
        self.comparison_parameters: T2 = None
        self.use_supplied_calcs_as_comp_inputs: bool = None
        self.notification_settings: DqNotificationSettings = None
        self.output_settings: DqOutputSettings = None
        self.dask_settings: DaskSettings = None



"""Includes PlotlyCommonFunctions class."""
from typing import TypeVar, Type

from plotly.graph_objs import Scatter, Table, Bar

from organon.afe.domain.reporting.abstract_report_helper import AbstractReportHelper
from organon.afe.domain.reporting.base_afe_output_report import BaseAfeOutputReport
from organon.afe.domain.reporting.plotly_dashboard_helper import get_line_scatter_plot, get_data_table_plot, \
    get_bar_chart, create_figure, figures_to_html, open_html_in_browser

ReportHelperType = TypeVar("ReportHelperType", bound=AbstractReportHelper)


class PlotlyCommonFunctions:
    """Common functions for plotly report helper classes."""

    @classmethod
    def generate_afe_report_dashboard(cls, report_helper_type: Type[ReportHelperType],
                                      afe_report: BaseAfeOutputReport):
        """Creates afe report dashboard"""
        fig1 = cls._get_afe_feature_dashboard_figure(report_helper_type, afe_report)
        fig2 = cls._get_resource_usage_dashboard_figure(afe_report)
        html = figures_to_html([fig1, fig2])
        open_html_in_browser(html)
        return [fig1, fig2]

    @classmethod
    def generate_afe_feature_dashboard(cls, report_helper_type: Type[ReportHelperType],
                                       afe_report: BaseAfeOutputReport):
        """Creates a table including afe output features information"""
        fig = cls._get_afe_feature_dashboard_figure(report_helper_type, afe_report)
        fig.show()

    @classmethod
    def _get_afe_feature_dashboard_figure(cls, report_helper_type: Type[ReportHelperType],
                                          afe_report: BaseAfeOutputReport):
        """Creates a table including afe output features information"""
        table = cls.__get_afe_feature_table(report_helper_type, afe_report)
        fig = create_figure([table], "Afe Features")
        return fig

    @classmethod
    def generate_coarse_class_dashboard(cls, afe_report: BaseAfeOutputReport, feature_name: str):
        """Creates bar plot showing transformation map corresponding to given feature"""
        bar_chart = cls.__get_coarse_class_bar_chart(afe_report, feature_name)
        fig = create_figure([bar_chart], "Coarse Class Dashboard", "Range", "Value")
        fig.show()

        return fig

    @classmethod
    def generate_memory_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of memory usage by date"""
        scatter = cls.__get_memory_usage_scatter(afe_report)
        fig = create_figure([scatter], "Memory Usage", "Date", "Usage(MB)")
        cls.__add_event_lines_to_figure(afe_report, fig)
        fig.show()
        return fig

    @classmethod
    def generate_cpu_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plot of cpu usage by date"""
        scatter = cls.__get_cpu_usage_scatter(afe_report)
        fig = create_figure([scatter], "CPU Usage", "Date", "Usage(percentage)")
        cls.__add_event_lines_to_figure(afe_report, fig, for_cpu=True)
        fig.show()
        return fig

    @classmethod
    def generate_resource_usage_dashboard(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plots of memory and cpu usages by date"""
        fig = cls._get_resource_usage_dashboard_figure(afe_report)
        fig.show()

    @classmethod
    def _get_resource_usage_dashboard_figure(cls, afe_report: BaseAfeOutputReport):
        """Creates a line-scatter plots of memory and cpu usages by date"""
        ts1 = cls.__get_memory_usage_scatter(afe_report)
        ts2 = cls.__get_cpu_usage_scatter(afe_report)

        fig = create_figure([ts1, ts2], "Resource Usages", "Date", ["Memory Usage(MB)", "CPU Usage(percentage)"])
        cls.__add_event_lines_to_figure(afe_report, fig)

        return fig

    @classmethod
    def __get_coarse_class_bar_chart(cls, afe_report: BaseAfeOutputReport, feature_name: str) -> Bar:
        """Returns a plotly Bar object generated for afe transformation map of given feature"""
        _map = afe_report.transformations[feature_name].tabulate_as_text()
        limits = list(_map.keys())
        map_vals = [float(i) for i in list(_map.values())]

        bar_chart = get_bar_chart(limits, map_vals)
        return bar_chart

    @classmethod
    def __get_afe_feature_table(cls, report_helper_type: Type[ReportHelperType],
                                afe_report: BaseAfeOutputReport) -> Table:
        """Returns a plotly Table object storing afe output features information"""
        data_frame = report_helper_type.get_output_features_as_data_frame(afe_report)
        table = get_data_table_plot(data_frame)
        return table

    @classmethod
    def __get_memory_usage_scatter(cls, afe_report: BaseAfeOutputReport) -> Scatter:
        """Returns a plotly Scatter object generated for memory usage by date"""
        usage = afe_report.memory_usage
        scatter = get_line_scatter_plot(list(usage.keys()), list(usage.values()), "Memory Usage")
        return scatter

    @classmethod
    def __get_cpu_usage_scatter(cls, afe_report: BaseAfeOutputReport) -> Scatter:
        """Returns a plotly Scatter object generated for cpu usage by date"""
        usage = afe_report.cpu_usage

        return get_line_scatter_plot(list(usage.keys()), list(usage.values()), "CPU Usage")

    @classmethod
    def __add_event_lines_to_figure(cls, afe_report, fig, for_cpu=False):
        """Adds vertical lines for events in plot"""
        tick_vals = list(afe_report.event_times.values())
        tick_texts = list(afe_report.event_times.keys())
        values_dict = afe_report.memory_usage
        if for_cpu:
            values_dict = afe_report.cpu_usage
        max_val = max(values_dict.values())
        total_time = max(values_dict.keys()) - min(values_dict.keys())
        for i, val in enumerate(tick_vals):
            if i < len(tick_vals) - 1:
                if (tick_vals[i + 1] - val).total_seconds() < total_time.total_seconds() / 100:
                    # prevents annotation overlapping by eliminating steps with a low execution time
                    continue
            fig.add_vline(x=val, line_dash="dot")
            fig.add_annotation(x=val, text=tick_texts[i], y=max_val, textangle=-90)



"""Includes DqNotificationSettings class."""
from organon.idq.domain.settings.mail_notification_settings import MailNotificationSettings


class DqNotificationSettings:
    """Notification settings for DQ execution."""
    def __init__(self):
        self.mail_notification_settings: MailNotificationSettings = None



"""Includes VotingEnsembleClassifier class."""
from typing import Dict, Any, List, Union

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_classifier import BaseClassifier
from organon.ml.modelling.algorithms.helpers.ensembling_helper import get_voting_classifier_prediction


class VotingEnsembleClassifier(BaseClassifier):
    """VotingEnsembleClassifier"""

    # pylint: disable=no-member

    def _predict_proba(self, data: pd.DataFrame) -> pd.DataFrame:
        estimators: List[BaseClassifier] = self.estimators
        columns = estimators[0].classes_
        prediction_dfs = (estimator.predict_proba(data).loc[:, columns] for estimator in estimators)  # generator
        average_probas = sum(prediction_dfs) / len(estimators)

        return average_probas

    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        estimators: List[BaseClassifier] = self.estimators
        if estimators is None or len(estimators) == 0:
            raise ValueError("Estimators not given")
        for estimator in estimators:
            if not isinstance(estimator, BaseClassifier):
                raise ValueError("All estimators should be an instance of BaseClassifier")
        if not self.prefit:
            for estimator in estimators:
                estimator.fit(train_data, target_data)

    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        estimators: List[BaseClassifier] = self.estimators
        all_pred_df = pd.DataFrame()
        target_col_name = None
        for i, estimator in enumerate(estimators):
            pred_df = estimator.predict(data)
            target_col_name = pred_df.columns.tolist()[0]
            all_pred_df[f"pred_{i}"] = pred_df.iloc[:, 0]
        final_pred = get_voting_classifier_prediction(all_pred_df)
        return pd.DataFrame({target_col_name: final_pred})

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        return {
            "estimators": None,
            "prefit": False,
        }

    def _get_classes(self) -> List[str]:
        return self.estimators[0].classes_  # predict_proba should return df with this column order too



"""Includes PretrainedModelType enum class"""
from enum import Enum, auto


class PretrainedModelType(Enum):
    """Pretrained model types"""
    XCEPTION = auto()
    RES_NET_50_V2 = auto()
    EFFICIENT_NET_B4 = auto()
    EFFICIENT_NET_B7 = auto()
    INCEPTION_RES_NET_V2 = auto()



"""Includes ReporterSettings class."""
from dataclasses import dataclass

import pandas as pd

from organon.ml.common.enums.target_type import TargetType


@dataclass
class ReporterSettings:
    """Settings for Reporter services"""
    data: pd.DataFrame
    target_column: str
    score_column: str
    target_type: TargetType
    id_str_column: str
    split_column: str
    num_bins: int



"""Includes Reporter class."""
from typing import Union, List

import numpy as np
import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.reporting.domain.objects.binary_report import BinaryReport
from organon.ml.reporting.domain.objects.multiclass_report import MulticlassReport
from organon.ml.reporting.domain.objects.regression_report import RegressionReport
from organon.ml.reporting.domain.services.base_reporter_service import BaseReporterService
from organon.ml.reporting.domain.services.binary_reporter_service import BinaryReporterService
from organon.ml.reporting.domain.services.multiclass_reporter_service import MulticlassReporterService
from organon.ml.reporting.domain.services.regression_reporter_service import RegressionReporterService
from organon.ml.reporting.settings.objects.multiclass_reporter_settings import MultiClassReporterSettings
from organon.ml.reporting.settings.objects.reporter_settings import ReporterSettings


class Reporter:
    """Modelling performance reporter"""

    def __init__(self, data: pd.DataFrame, target_column: str, score_column: str, target_type: str,
                 id_str_column: str = None, split_column: str = None, num_bins: int = 10,
                 probability_values: np.array = None, ordered_class_names: List = None):
        # pylint: disable=too-many-arguments
        target_type_enum = get_enum(target_type, TargetType)
        if target_type_enum == TargetType.MULTICLASS:
            self._settings = MultiClassReporterSettings(data, target_column, score_column, target_type_enum,
                                                        id_str_column, split_column, num_bins, probability_values,
                                                        ordered_class_names)
        else:
            self._settings = ReporterSettings(data, target_column, score_column, target_type_enum,
                                              id_str_column, split_column, num_bins)

        self.report: Union[BinaryReport, MulticlassReport, RegressionReport] = None

    def execute(self):
        """Generates and returns report"""
        service = self._get_reporter_service()
        self.report = service.execute(self._settings)
        return self.report

    def _get_reporter_service(self) -> BaseReporterService:
        if self._settings.target_type == TargetType.BINARY:
            return BinaryReporterService()
        if self._settings.target_type == TargetType.MULTICLASS:
            return MulticlassReporterService()
        return RegressionReporterService()



"""
This module includes base class for afe report helpers.
"""
import abc
from typing import Dict, TypeVar, Generic, List

import pandas as pd


from organon.afe.domain.common.persist_helper import PersistHelper
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.reporting.base_afe_output_report import FeatureCountReport, BaseAfeOutputReport

AfeModelOutputType = TypeVar("AfeModelOutputType", bound=BaseAfeModelOutput)
AfeOutputReportType = TypeVar("AfeOutputReportType", bound=BaseAfeOutputReport)


class AbstractReportHelper(Generic[AfeModelOutputType, AfeOutputReportType], metaclass=abc.ABCMeta):
    """
    This class is base abstract class for afe reporting.
    """

    @classmethod
    def __get_count_by_quantity_col(cls, output_features: Dict[str, BaseAfeFeature]):
        count_dict = {}
        for feature in output_features.values():
            qty_name = feature.afe_column.quantity_name
            if qty_name in count_dict:
                count_dict[qty_name] += 1
            else:
                count_dict[qty_name] = 1
        return count_dict

    @classmethod
    def __get_count_by_dimension_col(cls, output_features: Dict[str, BaseAfeFeature]):
        count_dict = {}
        for feature in output_features.values():
            dimension = feature.afe_column.dimension_name
            if dimension in count_dict:
                count_dict[dimension] += 1
            else:
                count_dict[dimension] = 1
        return count_dict

    @classmethod
    def __get_count_by_operator(cls, output_features: Dict[str, BaseAfeFeature]):
        count_dict = {}
        for feature in output_features.values():
            op_name = feature.afe_column.operator.name
            if op_name in count_dict:
                count_dict[op_name] += 1
            else:
                count_dict[op_name] = 1
        return count_dict

    @classmethod
    def _get_count_by_time_window(cls, output_features: Dict[str, BaseAfeFeature]):
        count_dict = {}
        for feature in output_features.values():
            time_window = feature.afe_column.time_window
            val = str(time_window)
            if val in count_dict:
                count_dict[val] += 1
            else:
                count_dict[val] = 1
        return count_dict

    @classmethod
    def _get_feature_counts_report(cls, output_features: Dict[str, BaseAfeFeature]):
        by_dim = AbstractReportHelper.__get_count_by_dimension_col(output_features)
        by_qty = AbstractReportHelper.__get_count_by_quantity_col(output_features)
        by_op = AbstractReportHelper.__get_count_by_operator(output_features)
        by_tw = AbstractReportHelper._get_count_by_time_window(output_features)
        return FeatureCountReport(by_dim, by_qty, by_op, by_tw)

    @classmethod
    def _get_dimension_names_map(cls, model_output: AfeModelOutputType):
        features = model_output.all_features if model_output.all_features is not None else model_output.output_features
        all_dimensions = [feature.afe_column.dimension_name for feature in features.values()]
        names_map = {}
        for dim in all_dimensions:
            names_map[dim] = model_output.transaction_file_stats.get_histogram(dim).reverse_index
        return names_map

    @classmethod
    @abc.abstractmethod
    def generate_afe_report(cls, model_output: AfeModelOutputType) -> AfeOutputReportType:
        """Creates output report from model output and runtime statistics"""

    @classmethod
    @abc.abstractmethod
    def generate_afe_report_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates afe report dashboard"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def generate_afe_feature_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates a table including afe output features information"""
        raise NotImplementedError

    @classmethod
    def generate_feature_extended_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates extended feature report"""
        raise NotImplementedError

    @classmethod
    def generate_feature_importance_dashboard(cls, afe_report: AfeOutputReportType, target_index: int = None,
                                              first_x_feature: int = None):
        """Creates afe feature importance dashboard"""
        raise NotImplementedError

    @classmethod
    def generate_roc_curve_dashboard(cls, afe_report: AfeOutputReportType, target_index: int = None, target_class=None):
        """Creates roc curve plot for features in given target"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def generate_coarse_class_dashboard(cls, afe_report: AfeOutputReportType, feature_name: str):
        """Creates bar plot showing transformation map corresponding to given feature"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def generate_memory_usage_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates a line-scatter plot of memory usage by date"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def generate_cpu_usage_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates a line-scatter plot of cpu usage by date"""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def generate_resource_usage_dashboard(cls, afe_report: AfeOutputReportType):
        """Creates a line-scatter plots of memory and cpu usages by date"""
        raise NotImplementedError

    @classmethod
    def get_output_features_as_data_frame(cls, afe_report: AfeOutputReportType):
        """Creates and returns a data frame of output features information"""
        features = list(afe_report.output_features.values())
        return cls._get_features_as_df(afe_report.dimension_names_map, features)

    @classmethod
    def _get_features_as_df(cls, dimension_names_map: Dict[str, Dict[int, str]], features: List[BaseAfeFeature]):
        col_info = []
        for feature in features:
            afe_col = feature.afe_column
            dimension = feature.afe_column.dimension_name
            dimension_names_dict = dimension_names_map[dimension] if dimension in dimension_names_map else {}
            meaning = cls.get_feature_meaning(feature, dimension_names_dict)
            tw_str = cls._get_time_window_string(feature)
            col_info.append(
                {
                    "Feature": feature.feature_name,
                    "Dimension": afe_col.dimension_name,
                    "Quantity": afe_col.quantity_name,
                    "Operator": afe_col.operator.name,
                    "Date Column": afe_col.date_column.column_name if afe_col.date_column is not None else "",
                    "Time Window": tw_str,
                    "Resolution": afe_col.date_resolution.name,
                    "Meaning": meaning
                }
            )

        data_frame = pd.DataFrame(col_info)
        data_frame.fillna('', inplace=True)

        return data_frame

    @classmethod
    def _get_time_window_string(cls, afe_feature:BaseAfeFeature):
        return str(afe_feature.afe_column.time_window)

    @classmethod
    def get_feature_meaning(cls, feature: BaseAfeFeature, group_names: Dict[int, str]) -> str:
        """Returns meaning of given afe output feature"""
        return PersistHelper.get_feature_meaning(feature, group_names)



"""Includes DqControlParameters class."""
from typing import List, Dict

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_test_group import DqTestGroup
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType
from organon.idq.domain.settings.abstractions.dq_base_calculation_parameters import DqBaseCalculationParameters
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters
from organon.idq.domain.settings.dq_comparison_column_info import DqComparisonColumnInfo


class DqControlParameters:
    """Parameters for DqControl execution"""

    def __init__(self):
        self.comparison_parameters: DqBaseComparisonParameters = None
        self.control_parameters: List[DqBaseCalculationParameters] = None
        self.control_results: List[DqCalculationResult] = None
        self.test_calc_parameters: DqBaseCalculationParameters = None
        self.test_calc_result: DqCalculationResult = None
        self.test_group_info: Dict[DqTestGroupType, DqTestGroup] = None

    def get_comparison_cols_by_native_type(self, col_native_type: ColumnNativeType) -> List[DqComparisonColumnInfo]:
        """Returns comparison column info for columns of given native type"""
        comparison_cols = self.comparison_parameters.comparison_columns
        data_col_collection = self.test_calc_result.data_source_stats.data_column_collection
        col_names = [col.column_name for col in data_col_collection if col.column_native_type == col_native_type]
        return [col for col in comparison_cols if col.column_name in col_names]

    def get_nominal_comparison_columns(self) -> List[DqComparisonColumnInfo]:
        """Returns nominal comparison columns"""
        nominal_columns = []
        for col_type in self.test_calc_parameters.nominal_column_types:
            nominal_columns.extend(self.get_comparison_cols_by_native_type(col_type))
        return nominal_columns



"""
This module keeps functions performing the string operations.
"""


def get_bytes(string: str) -> bytearray:
    """
    Returns the string as a byte array.
    :param string: str
    :return: bytearray
    """
    return bytearray(string, "utf-8")


def is_alpha_numeric(string: str) -> bool:
    """
    Checks if the string is an alpha-numeric string.
    :param string: str
    :return: bool
    """
    return string.isalnum()


def is_word(string: str) -> bool:
    """
    Checks if the string is an alphabetic string.
    :param string: str
    :return: bool
    """
    return string.isalpha()


def is_number(string: str) -> bool:
    """
    Checks if the string is an numeric string.
    :param string: str
    :return: bool
    """
    return string.isnumeric()


def get_db_column_name_compatible_string(input_str: str, max_length: int) -> str:
    """
    Filters the input string according to the valid char list and returns it.
    :param input_str: str
    :param max_length: int
    :return: str
    """
    i: int = 0
    tot_chars: int = 0
    valid_char_list: str = "0123456789abcdefghjklmnoprqstuxwvyzABCDEFGHJKLMNOPRQSTUXWVYZ_"
    output_str: str = ""

    while tot_chars < max_length and i < len(input_str):

        char: str = input_str[i]
        if char in valid_char_list:
            output_str += char
            tot_chars += 1

        i += 1

    return output_str


def is_empty(string: str) -> bool:
    """
    Checks if the string is empty.
    :param string: str
    :return: bool
    """
    return not string


def remove_single_character(string: str, char: str) -> str:
    """
    Replaces a character from a string and returns it.
    :param string: str
    :param char: str
    :return: str
    """
    return string.replace(char, "")


def contains_white_space(string: str) -> bool:
    """
    Checks if the string contains white space.
    :param string: str
    :return: bool
    """
    return " " in string


def nullable_length(input_str: str = None) -> str:
    """
    Checks the input string is none or empty.
    If it is none or empty string, return "null", if it is not returns the length of the string.
    :param input_str: str
    :return: str
    """
    if input_str is None or input_str == "":
        return "null"
    return str(len(input_str))


def effective_value(input_str: str = None) -> str:
    """
    Checks the input string is none or empty.
    If it is none returns "null"
    If it is empty returns "empty"
    If both of them are not correct return the input string.
    :param input_str: str
    :return: str
    """
    if input_str is None:
        return "null"
    if input_str.strip() == "":
        return "empty"
    return input_str


def equals_ignoring_case(input_str: str, compared_string: str) -> bool:
    """
    Compares the two strings with ignoring the case.
    :param input_str: str
    :param compared_string: str
    :return: bool
    """
    if input_str is None and compared_string is None:
        return True

    if input_str is None or compared_string is None:
        return False

    val = input_str.upper() == compared_string.upper()

    return val


def to_upper_eng(input_str: str) -> str:
    """
    Returns the upper case representation of the string.
    :param input_str: str
    :return: str
    """
    val = input_str.upper()
    return val


def to_lower_eng(input_str: str) -> str:
    """
    Returns the lower case representation of the string.
    :param input_str: str
    :return: str
    """
    val = input_str.lower()
    return val



"""Includes FeatureSimilaritySelectionService class."""
from typing import List

import numpy as np
import pandas as pd

from organon.fl.core.helpers.data_frame_helper import get_correlation_matrix_memory_efficient
from organon.fl.mathematics.constants import DOUBLE_MAX
from organon.ml.feature_selection.domain.objects.feature_similarity_selection_output import \
    FeatureSimilaritySelectionOutput
from organon.ml.feature_selection.domain.objects.settings.feature_similarity_selection_settings import \
    FeatureSimilaritySelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import SelectionRunSettings
from organon.ml.feature_selection.domain.services.base_unsupervised_feature_selection_service import \
    BaseUnsupervisedFeatureSelectionService


class FeatureSimilaritySelectionService(BaseUnsupervisedFeatureSelectionService[FeatureSimilaritySelectionSettings,
                                                                                FeatureSimilaritySelectionOutput]):
    """Service class for Feature Similarity selection"""

    @classmethod
    def _validate_settings(cls, settings: FeatureSimilaritySelectionSettings):
        pass

    @classmethod
    def _run(cls, run_settings: SelectionRunSettings) -> FeatureSimilaritySelectionOutput:
        settings: FeatureSimilaritySelectionSettings = run_settings.settings
        data_frame = run_settings.data
        output = FeatureSimilaritySelectionOutput()
        variances = cls._get_column_variances(data_frame)
        zero_variance_cols = list((variances[variances == 0]).index)
        nan_max_cols = [col for col in data_frame if np.isnan(data_frame[col].max())]
        dropped_columns = set(zero_variance_cols + nan_max_cols)
        col_list = data_frame.columns.difference(dropped_columns).to_list()
        if len(col_list) <= 1:
            output.selected_features = col_list
            return output

        mci_df = cls.__get_corr_df(data_frame, variances, col_list, settings.prioritize_performance)

        k_val = round(len(mci_df.columns) * settings.selection_percent)
        if k_val > mci_df.shape[0] - 1:
            k_val = mci_df.shape[0] - 1

        if k_val > 1:
            min_rik = cls.__get_min_rik_and_update_corr_df(mci_df, k_val)
            eps = min_rik
            while True:
                if k_val > mci_df.shape[0] - 1:
                    k_val = mci_df.shape[0] - 1
                if k_val <= 1:
                    break
                k_val = k_val - 1
                min_rik = cls.__get_min_rik_and_update_corr_df(mci_df, k_val)
                if np.isnan(min_rik) or min_rik <= eps:
                    break

        selected_set = mci_df.columns
        output.selected_features = list(selected_set)
        return output

    @classmethod
    def _get_column_variances(cls, data_frame: pd.DataFrame):
        # calculate variances column by column for memory efficiency
        variances = []
        for col in data_frame:
            variances.append(data_frame[col].var())
        return pd.Series(variances, index=data_frame.columns)

    @classmethod
    def __get_min_rik_and_update_corr_df(cls, mci_df: pd.DataFrame, k_val: int):
        min_rik_feature, min_rik = cls.__find_min_rik(mci_df, k_val)
        discard_list = list(mci_df[min_rik_feature].sort_values().index)[:k_val]
        mci_df.drop(discard_list, axis=1, inplace=True)
        mci_df.drop(discard_list, axis=0, inplace=True)
        return min_rik

    @classmethod
    def __find_min_rik(cls, mci_df: pd.DataFrame, k_val: int):
        min_rik_feature = None
        min_rik = DOUBLE_MAX
        for col in mci_df.columns:
            kth_col = list(mci_df[col].sort_values().index)[k_val - 1]
            rik = mci_df.loc[col, kth_col]
            if rik < min_rik:
                min_rik_feature = col
                min_rik = rik
        return min_rik_feature, min_rik

    @classmethod
    def __get_corr_df(cls, data_frame: pd.DataFrame, variances: pd.Series, col_list: List[str],
                      prioritize_performance: bool):
        mci_df = cls._get_correlation_matrix(data_frame, col_list, prioritize_performance)
        same_drop_cols = set()
        for col1_index, col1 in enumerate(col_list):
            mci_df.loc[col1, col1] = np.nan
            if col1_index == len(col_list):
                break
            var1 = variances[col1]
            other_cols = col_list[col1_index + 1:]
            new_vals = []
            for col2 in other_cols:
                corr = mci_df.loc[col1, col2]
                var2 = variances[col2]
                mci = 0.5 * (var1 + var2 - np.sqrt(
                    np.square((var1 + var2)) - 4 * var1 * var2 * (1 - np.square(corr))))
                new_vals.append(mci)
                if mci == 0:
                    same_drop_cols.add(col2)
            mci_df.loc[col1, other_cols] = new_vals
            mci_df.loc[other_cols, col1] = new_vals
        mci_df.drop(same_drop_cols, axis=1, inplace=True)
        mci_df.drop(same_drop_cols, axis=0, inplace=True)

        drop_list = list(mci_df[mci_df.max().isnull()].index)
        mci_df.drop(drop_list, axis=1, inplace=True)
        mci_df.drop(drop_list, axis=0, inplace=True)
        return mci_df

    @classmethod
    def _get_correlation_matrix(cls, data: pd.DataFrame, col_list: List[str], prioritize_performance: bool):
        if prioritize_performance:
            return data.corr().loc[col_list, col_list]
        return get_correlation_matrix_memory_efficient(data, columns=col_list)



"""Includes SampleDataServiceHelper class."""
from datetime import datetime
from typing import Dict, Tuple, Callable

import pandas as pd

from organon.afe.domain.enums.record_source_type import RecordSourceType
from organon.afe.domain.settings.auto_column_decider_settings import AutoColumnDeciderSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval
from organon.fl.core.exceptionhandling.known_exception import KnownException


class SampleDataServiceHelper:
    """Includes common methods for sample data services."""

    def __init__(self, trx_record_source: RecordSource,
                 entity_col: str, func_per_date_column: Dict[str, Callable[[str], datetime]],
                 interval_per_entity_per_date_column: Dict[str, Dict[str, DateInterval]],
                 distinct_entity_source: RecordSource, auto_col_decider_settings: AutoColumnDeciderSettings):
        self.trx_record_source = trx_record_source
        self.__trx_entity_column = entity_col
        self.func_per_date_column = func_per_date_column
        self.interval_per_entity_per_date_column = interval_per_entity_per_date_column
        self.distinct_entity_source = distinct_entity_source
        self.auto_column_decider_settings = auto_col_decider_settings

    def read_dataframe_source(self, unique_entities_set, min_max_dates_per_date_column):
        """
        Reads pandas DataFrame source.
        :return: TransactionFileRecordCollection containing DataFrame records
        """
        if self.distinct_entity_source is not None and \
                self.distinct_entity_source.get_type() != RecordSourceType.DATA_FRAME:
            raise KnownException("Distinct entity source should be a dataframe")
        data_frame: pd.DataFrame = self.trx_record_source.source

        return self._get_reduced_frame(data_frame, min_max_dates_per_date_column, unique_entities_set)

    def read_text_source(self, unique_entities_set, min_max_dates_per_date_column, chunksize):
        """
        Reads text source.
        :return: TransactionFileRecordCollection containing text records
        """
        if self.distinct_entity_source is not None and \
                self.distinct_entity_source.get_type() != RecordSourceType.DATA_FRAME:
            raise KnownException("Distinct entity source should be a dataframe")

        record_source = self.trx_record_source
        frame_chunks = []
        for data_frame in pd.read_csv(record_source.source, sep=",", chunksize=chunksize):
            reduced_frame = self._get_reduced_frame(data_frame, min_max_dates_per_date_column, unique_entities_set)
            frame_chunks.append(
                reduced_frame.sample(
                    frac=self.auto_column_decider_settings.sampling_ratio))

        return pd.concat(frame_chunks)

    def _get_reduced_frame(self, data_frame: pd.DataFrame,
                           min_max_dates_per_date_column: Dict[str, Tuple[datetime, datetime]],
                           unique_entities_set: set = None):
        entity_ids = data_frame[self.__trx_entity_column].astype(str)
        all_conditions_series = None
        for date_col in min_max_dates_per_date_column:
            dates = data_frame[date_col].astype(str).map(self.func_per_date_column[date_col])
            min_date, max_date = min_max_dates_per_date_column[date_col]
            condition_series = (max_date >= dates) & (dates >= min_date) & \
                               entity_ids.isin(self.interval_per_entity_per_date_column[date_col])
            if all_conditions_series is None:
                all_conditions_series = condition_series
            else:
                all_conditions_series |= condition_series
        if all_conditions_series is None:
            return data_frame

        if unique_entities_set:
            all_conditions_series &= entity_ids.isin(unique_entities_set)

        return data_frame.loc[all_conditions_series]



"""Includes DqTestType class."""


class DqTestType:
    """Dq test types"""
    TABLE_NEW_FIELD_CONTROL = 1
    TABLE_REMOVED_FILED_CONTROL = 2
    TABLE_FILED_TYPE_CONTROL = 3
    TABLE_FILED_LENGTH_CONTROL = 4
    TABLE_EMPTY_TABLE_CONTROL = 5
    TABLE_SIZE_CI_CONTROL = 6
    TABLE_SIZE_MODELLING_AND_CI_CONTROL = 7
    TABLE_SIZE_MODELLING_AND_TL_CONTROL = 8
    TABLE_SIZE_MODELLING_AND_AD_CONTROL = 9
    TABLE_COUNT_CI_CONTROL = 10
    TABLE_COUNT_MODELLING_AND_CI_CONTROL = 11
    TABLE_COUNT_MODELLING_AND_TL_CONTROL = 12
    TABLE_COUNT_MODELLING_AND_AD_CONTROL = 13
    TABLE_SIZE_TL_CONTROL = 14
    TABLE_COUNT_TL_CONTROL = 15
    PSI = 16
    TRAFFIC_LIGHT = 17
    UNEXPECTED_VALUE_CONTROL = 18
    MEAN_TL = 19
    MEAN_CI = 20
    MEAN_AD = 21
    MODELLING_AND_TL = 22
    MODELLING_AND_CI = 23
    COLUMN_RANGE_CONTROL = 24
    COLUMN_POPULATION_PERCENT_CONTROL = 25
    COLUMN_VALID_VALUE_CONTROL = 26
    COLUMN_UNVALID_VALUE_CONTROL = 27
    CARDINALITY_CONTROL = 28
    ALL_VALUES_ARE_ZERO_CONTROL = 29
    ALL_VALUES_ARE_NULL_CONTROL = 30



"""
This module includes BaseAfeScoringSettings class.
"""
from datetime import datetime
from typing import TypeVar, Generic

from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.settings.afe_process_settings import AfeProcessSettings
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.base_afe_settings import BaseAfeSettings
from organon.afe.domain.settings.record_source import RecordSource

AfeModelOutputType = TypeVar("AfeModelOutputType", bound=BaseAfeModelOutput)


class BaseAfeScoringSettings(Generic[AfeModelOutputType], BaseAfeSettings):
    """
    #TODO docstring
    """
    ATTR_DICT = {
        "process_settings": AfeProcessSettings,
        "model_file_path": str,
        "scoring_date": datetime,
        "target_record_source": RecordSource,
        "raw_input_source": RecordSource,
        "trx_reading_settings": AfeDataReadingSettings,
    }

    def __init__(self):
        super().__init__()
        self.model_file_path: str = None
        self.scoring_date: datetime = None
        self.target_record_source: RecordSource = None
        self.raw_input_source: RecordSource = None
        self.model_output: BaseAfeModelOutput = None
        self.trx_reading_settings: AfeDataReadingSettings = None



"""todo"""
from enum import Enum


class DqTestGroupType(Enum):
    """todo"""
    TABLE_SCHEMA_CONTROLS = 1
    RULE_BASED_CONTROLS_TABLE_SET = 2
    MODELLING_AND_CI_CONTROLS_TABLE_SET = 3
    TRAFFIC_LIGHT_CONTROLS = 4
    DISTRIBUTIONAL_COMPARISONS = 5
    MODELLING_AND_CI_CONTROLS_COLUMN_SET = 6
    RULE_BASED_CONTROLS_COLUMN_SET = 7



"""Includes BaseUserCalculationParams class."""

from organon.idq.services.user_settings.user_input_source_settings import UserInputSourceSettings


class BaseUserCalculationParams:
    """Base calculation settings user class"""

    def __init__(self, calc_name, input_source_settings, max_nom_card_count, outlier_param,
                 use_pop_nom_stats):
        self.calculation_name: str = calc_name
        self.input_source_settings: UserInputSourceSettings = input_source_settings
        self.max_nominal_cardinality_count: int = max_nom_card_count
        self.outlier_parameter: float = outlier_param
        self.use_population_nominal_stats: bool = use_pop_nom_stats



"""Includes common methods in object classification domain"""
from organon.ml.object_classification.domain.objects.data_source_settings import DataSourceSettings
from organon.ml.object_classification.domain.objects.object_clf_settings import ObjectClfSettings


def get_data_source_settings(obj_clf_settings: ObjectClfSettings) -> DataSourceSettings:
    """Returns data source settings object for given Object Classification Settings"""
    return DataSourceSettings(obj_clf_settings.clf_mode, obj_clf_settings.validation_data_ratio,
                              obj_clf_settings.image_size, obj_clf_settings.batch_size, obj_clf_settings.color_mode,
                              obj_clf_settings.random_seed)



"""
This module includes TargetFileDescriptor class.
"""
from datetime import datetime

from organon.afe.domain.common.reader_helper import get_values_from_kwargs
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.afe_target_column import AfeTargetColumn


class TargetDescriptor:
    """
    Class for information about target file or database table
    """

    ATTR_DICT = {
        "entity_column_name": str,
        "default_measurement_date": datetime,
        "date_column": AfeDateColumn,
        "target_column": AfeTargetColumn,
        "reading_settings": AfeDataReadingSettings
    }

    def __init__(self, **kwargs):
        self.entity_column_name: str = None
        self.default_measurement_date: datetime = None
        self.date_column: AfeDateColumn = None
        self.target_column: AfeTargetColumn = None
        self.reading_settings: AfeDataReadingSettings = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""
This module includes TemporalGrid class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs


class TemporalGrid:
    """
    Class to define a horizon list for AFE algorithm settings
    """
    ATTR_DICT = {
        "length": int,
        "stride": int,
        "offset": int
    }

    def __init__(self, **kwargs):
        self.offset: int = None
        self.stride: int = None
        self.length: int = None

        get_values_from_kwargs(self, TemporalGrid.ATTR_DICT, kwargs)



"""Includes Dask DataFrame class for future development"""
from typing import List, Union, Any

from organon.fl.core.businessobjects.idataframe import IDataFrame


class DaskDataFrame(IDataFrame):
    """Dask DataFrame based custom DataFrame implementing IDataFrame abstract class."""

    def set_value(self, value, column_name: str, row_index: int):
        raise NotImplementedError()

    def contains_column(self, column_name: str) -> bool:
        raise NotImplementedError()

    def try_add(self, column_name: str, data: List[Any], mark_for_garbage_collection: bool = False):
        raise NotImplementedError()

    def remove(self, column_names: Union[List[str], str]):
        raise NotImplementedError()

    def collect_garbage_columns(self):
        raise NotImplementedError()

    def get_value(self, column_name: str, row_index: int = None):
        raise NotImplementedError()

    def get_column_names(self) -> List[str]:
        raise NotImplementedError()

    def get_subset_as_pandas_df(self, indices: List[Any] = None, columns: List[str] = None):
        raise NotImplementedError()

    def copy_as_pandas_df(self):
        raise NotImplementedError()



"""settings class for Grid Search Text Classification Models"""
from typing import List

from organon.ml.text_classification.domain.enums.model_checkpoints import ModelCheckpoints
from organon.ml.text_classification.domain.enums.selection_metrics import SelectionMetrics


class GridSearchSettings:
    """Class for grid search parameters"""

    def __init__(self, models: List[ModelCheckpoints] = None, batch_sizes: List[int] = None, epochs: List[int] = None,
                 learning_rates: List[float] = None, early_stopping_patiences: List[int] = None,
                 early_stopping_min_deltas: List[float] = None,
                 model_selection_metric: SelectionMetrics = SelectionMetrics.VAL_LOSS):
        # pylint: disable=too-many-arguments
        self.models = models if models is not None else [ModelCheckpoints.BERT_BASE,
                                                         ModelCheckpoints.BERT_BASE_MLINGUAL]
        self.batch_sizes = batch_sizes if batch_sizes is not None else [8]
        self.epochs = epochs if epochs is not None else [2]
        self.learning_rates = learning_rates if learning_rates is not None else [1e-5]
        self.early_stopping_patiences = early_stopping_patiences if early_stopping_patiences is not None else [2]
        self.early_stopping_min_deltas = early_stopping_min_deltas if early_stopping_min_deltas is not None else [1e-3]
        self.model_selection_metric = model_selection_metric



"""Module for DenseMatrix"""
import numpy as np
from organon.fl.mathematics.sweep.matrices.i_matrix import IMatrix


class DenseMatrix(IMatrix):
    """Matrix class for dense matrices"""

    @property
    def is_centralized(self):
        """Get the centralized information"""
        return self._is_centralized

    @property
    def data(self):
        """Get the data"""
        return self._data

    @property
    def shape(self):
        """Get the shape of the matrix"""
        return self._shape

    @property
    def means(self) -> np.ndarray:
        """Get the column means of the matrix"""
        return self._means

    def __init__(self, data: np.ndarray):
        if len(data.shape) != 2:
            raise ValueError("Array should be 2 dimensional")
        self._data = data
        self._shape = data.shape
        self._means = np.mean(data, axis=0)

        self.centralize_columns()
        self._check_centralized()

    def _check_centralized(self):
        """Check if the matrix centralized"""
        self._is_centralized = True

        for j in range(self.shape[1]):
            if ~np.isclose(np.mean(self._data[:, j]), 0):
                self._is_centralized = False

    def get_column(self, k: int):
        """Get specified column of the matrix"""
        return self._data[:, k].copy()

    def centralize_columns(self):
        """Centralize all columns"""
        self._data -= np.mean(self._data, axis=0)

        self._is_centralized = True

    def mul_vector(self, vector: np.ndarray):
        """Calculates Matrix Vector product"""
        return np.dot(self._data.T, vector)



"""Includes PsiNominalSignalControl class."""
from typing import List, Dict

from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.dq_constants import DqConstants
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.controls.base_psi_signal_control import BasePsiSignalControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.signal_type import SignalType
from organon.idq.domain.helpers.statistics.dq_statistical_functions import DqStatisticalFunctions


class PsiNominalSignalControl(BasePsiSignalControl):
    """Control class psi comparison for nominal column"""

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.PSI_NOMINAL_VALUES

    def get_description(self) -> str:
        return "Psi Nominal Signals"

    def _execute_control(self) -> List[DqComparisonResult]:
        results = []
        nominal_comp_cols = self.comp_context.get_nominal_comparison_columns()
        test_group_info = self.comp_context.test_group_info[self.get_test_group_type()]
        for col in nominal_comp_cols:
            col_name = col.column_name
            filtered_controls = self._get_filtered_columns(col, test_group_info)
            test_nominal_stat_collection = self.comp_context.test_calc_result.population_stats.nominal_statistics
            if test_nominal_stat_collection is None:
                LogHelper.info("Test statistics collection is null, psi tests skipped")

            elif len(filtered_controls) > 0 and col_name in test_nominal_stat_collection:
                test_nominal_stats = test_nominal_stat_collection[col_name]
                self.__control_executor(test_nominal_stats, filtered_controls, col_name, results)

        return results

    def __control_executor(self, test_nominal_stats: DqCategoricalStatistics,
                           filtered_controls: List[DqCalculationResult], col_name: str,
                           results: List[DqComparisonResult]):
        test_frequencies = test_nominal_stats.frequencies
        max_psi = 0.0
        maximum_psi_sample = ""
        maximum_psi_category = ""
        maximum_psi_control_set_frequencies: Dict[str, float] = {}
        psi_per_control_set: Dict[str, float] = {}
        for control in filtered_controls:
            stat_collection = control.population_stats.nominal_statistics
            if col_name not in stat_collection:
                continue
            nominal_stats = stat_collection[col_name]
            control_frequencies = nominal_stats.frequencies
            psi_tuple = DqStatisticalFunctions.population_stability_index(control_frequencies,
                                                                          test_frequencies,
                                                                          DqConstants.MINIMUM_INJECTION)
            psi_value = psi_tuple[1]
            sample_to_string = control.calculation_name
            psi_per_control_set[sample_to_string] = psi_value
            if psi_value > max_psi:
                max_psi = psi_value
                maximum_psi_sample = sample_to_string
                maximum_psi_category = psi_tuple[0]
                maximum_psi_control_set_frequencies = {str(key): value for (key, value) in
                                                       control_frequencies.items()}
        psi_signal = self.get_signal_type(max_psi)
        if psi_signal == SignalType.GREEN:
            return
        result_code = DqComparisonResultCode.PSI_RED_SIGNAL if psi_signal == SignalType.RED \
            else DqComparisonResultCode.PSI_YELLOW_SIGNAL
        self.check_maximum_psi(col_name, result_code, results, max_psi, "MaximumPsi")
        self.check_maximum_psi_control_set(col_name, result_code, results, maximum_psi_sample,
                                           "MaximumPsiControlSet")
        self.check_maximum_psi_category(col_name, result_code, results, maximum_psi_category,
                                        "MaximumPsiCategory")
        self.check_test_frequency(col_name, result_code, results, test_frequencies,
                                  "NominalColumnTestFrequencies")
        self.check_max_psi_control_frequency(col_name, result_code, results,
                                             maximum_psi_control_set_frequencies,
                                             "NominalColumnMaximumPsiControlFrequencies")
        self.check_psi_per_control_set(col_name, result_code, results, psi_per_control_set,
                                       "PsiPerControlSet")



"""
This module includes AfeOutputReport class.
"""

from organon.afe.domain.modelling.businessobjects.afe_feature import AfeFeature
from organon.afe.domain.reporting.base_afe_output_report import BaseAfeOutputReport


class AfeOutputReport(BaseAfeOutputReport[AfeFeature]):
    """Afe output information to be reported."""



"""Includes SelectionSettings class."""
from dataclasses import dataclass
from typing import List

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


@dataclass
class SelectionSettings:
    """Settings for model selection"""
    modellers: List[BaseModeller]
    train_data: pd.DataFrame
    target_data: pd.Series
    target_type: TargetType
    cv_count: int = None
    test_data: pd.DataFrame = None
    test_target_data: pd.Series = None
    add_stacking: bool = True
    add_voting: bool = True
    num_threads_for_parallel_fit: int = None
    num_threads_for_parallel_cv: int = None



"""Includes OneHotEncodingSettings class."""

from dataclasses import dataclass


@dataclass
class OneHotEncodingSettings:
    """Settings for OneHotEncoding service"""
    threshold: float
    max_bin: int



"""Includes TupleGenericCollection class."""
from typing import TypeVar, List, Generic, Tuple, Dict

from sortedcontainers import SortedDict

T = TypeVar("T")  # pylint: disable=invalid-name


class TupleGenericCollection(Generic[T]):
    """A collection storing value-tuple mappings. There is a maximum number(capacity) of tuples that can be stored."""

    def __init__(self, capacity: int):
        self.__capacity: int = capacity
        self.__list_per_tuple: Dict[Tuple, List[T]] = SortedDict()
        self.__tuple_reverse_list: Dict[T, Tuple] = SortedDict()

    def try_add(self, _tuple: Tuple, datum: T) -> bool:
        """Adds given _tuple-datum mapping to collection if capacity is not exceeded."""
        if _tuple not in self.__list_per_tuple:
            if len(self.__list_per_tuple) >= self.__capacity:
                return False
            self.__list_per_tuple[_tuple] = []
        self.__list_per_tuple[_tuple].append(datum)
        self.__tuple_reverse_list[datum] = _tuple
        return True

    @property
    def count(self):
        """Returns number of tuple mappings in collection."""
        return len(self.__list_per_tuple)

    @property
    def capacity(self):
        """Returns capacity of collection."""
        return self.__capacity

    def __getitem__(self, datum: T):
        return self.__tuple_reverse_list[datum]

    @property
    def list_per_tuple(self):
        """Returns tuple-value_list mapping in collection."""
        return self.__list_per_tuple



"""Includes DataSourceStatsTrafficLightControl class."""
from typing import List

from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.algorithms.objects.mean_ci_comparison_input import MeanConfidenceIntervalComparisonInput
from organon.idq.domain.algorithms.objects.traffic_light_comparison_input import TrafficLightComparisonInput
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.controls.helpers.traffic_light_comparer import TrafficLightComparer
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class DataSourceStatsTrafficLightControl(BaseDqControl):
    """Control for data source statistics using Traffic Light """

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.TRAFFIC_LIGHT_CONTROLS

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.DATA_SOURCE_STATS_TL_CONTROL

    def get_description(self) -> str:
        return "Row Count Traffic Light Control"

    def _execute_control(self) -> List[DqComparisonResult]:
        bmh = self.get_test_bmh()
        if bmh < 1:
            raise ValueError("Bmh should be greater than or equal to 1")
        row_count_series = self.__get_row_count_series()
        row_count_current_value = self.comp_context.test_calc_result.data_source_stats.row_count

        results = TrafficLightComparer.get_mi_tl_comparison_results(
            self.__get_ci_comparison_input(row_count_series, row_count_current_value),
            self.__get_tl_comparison_input(row_count_series, row_count_current_value)
        )
        return results

    def __get_tl_comparison_input(self, series: List[float],
                                  current_value: int) -> TrafficLightComparisonInput:
        inp = TrafficLightComparisonInput()
        inp.test_group = self.get_test_group_type()
        inp.data_entity = DataEntityType.TABLE
        inp.data_entity_name = self.comp_context.test_calc_parameters.input_source_settings.source.get_name()
        inp.past_series = series
        inp.current_value = current_value
        inp.green_threshold = self.comp_context.comparison_parameters.traffic_light_threshold_green
        inp.yellow_threshold = self.comp_context.comparison_parameters.traffic_light_threshold_yellow
        inp.result_code = DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_TRAFFIC_LIGHT_INTERVAL
        inp.message = "Mean value for the historical row-count is statistically different from current row-count"
        return inp

    def __get_ci_comparison_input(self, series: List[float], current_value: int) \
            -> MeanConfidenceIntervalComparisonInput:
        inp = MeanConfidenceIntervalComparisonInput()
        inp.test_group = self.get_test_group_type()
        inp.data_entity = DataEntityType.TABLE
        inp.data_entity_name = self.comp_context.test_calc_parameters.input_source_settings.source.get_name()
        inp.past_series = series
        inp.current_value = current_value
        inp.z_score = self.comp_context.comparison_parameters.z_score
        inp.result_code = DqComparisonResultCode.ROW_COUNT_MEAN_IS_OUTSIDE_MEAN_CONFIDENCE_INTERVAL
        inp.message = "Mean value for the historical row-count series is statistically different from current row-count"
        return inp

    def __get_row_count_series(self) -> List[float]:
        series = []
        bmh = self.get_test_bmh()
        control_results = self.comp_context.control_results
        row_count_eligible = 0
        for i in range(len(control_results) - 1, -1, -1):
            row_count = control_results[i].data_source_stats.row_count
            if row_count is not None and row_count_eligible < bmh:
                series.append(float(row_count))
                row_count_eligible += 1
        return list(reversed(series))



"""Includes OneHotEncodingService class"""
from typing import List, Dict, Optional

import numpy as np
import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.fl.mathematics.constants import INT_MAX
from organon.ml.preprocessing.settings.objects.one_hot_encoding_settings import OneHotEncodingSettings


class OneHotEncodingService:
    """Service for one hot encoding"""

    NULLS_COL_KEY = "_$NullValue$_"
    OTHERS_COL_KEY = "_$OtherValues$_"

    def __init__(self, settings: OneHotEncodingSettings):
        self._settings = settings
        self._categorical_cols = None
        self._features_dict: Dict[str, List[str]] = None
        self._columns_dict: Dict[str, List[str]] = None
        self._all_ohe_columns: Dict[str, Dict[str, str]] = None

    @property
    def features_dict(self):
        """Returns selected features per categorical column"""
        return self._features_dict

    @property
    def columns_dict(self):
        """Returns generated columns per categorical columns (after transform)"""
        return self._columns_dict

    def fit(self, data: pd.DataFrame):
        """Fit encoder"""
        if self._settings.threshold < 0:
            raise ValueError("Threshold cannot be negative.")
        if self._settings.max_bin < 0:
            raise ValueError("Maximum number of features cannot be negative.")
        if data is None or data.empty:
            raise ValueError("Train data cannot be null or empty.")
        self._categorical_cols = [col for col in data if get_column_native_type(data, col) == ColumnNativeType.String]
        self._features_dict = {}
        for col in self._categorical_cols:
            self._features_dict[col] = self._get_col_features(data, col, self._settings.threshold,
                                                              self._settings.max_bin)
        return self

    def transform(self, data: pd.DataFrame, inplace: bool = False):
        """Transform data"""
        if self._features_dict is None:
            raise ValueError("Encoder not fitted yet")
        if data is None or data.empty:
            raise ValueError("Transformation data cannot be null or empty.")
        if not inplace:
            trans_data = data.copy()
        else:
            trans_data = data
        different_columns = [col for col in trans_data if
                             (get_column_native_type(trans_data, col) == ColumnNativeType.String and
                              col not in self._categorical_cols)]
        if different_columns:
            raise ValueError(
                f"These columns are not found in the train data or"
                f" their types were not categorical: {', '.join(different_columns)}.")

        generated_cols_per_category = {}
        all_ohe_col_names_dict = {}
        all_ohe_columns = []
        data_cols = trans_data.columns.tolist()
        for col in self._categorical_cols:
            features = self._features_dict[col]
            new_col_names_dict = self._get_all_new_column_names(data_cols + all_ohe_columns, col, features)
            new_col_names = list(new_col_names_dict.values())
            all_ohe_columns.extend(new_col_names)
            generated_cols_per_category[col] = new_col_names
            all_ohe_col_names_dict[col] = new_col_names_dict
        self._all_ohe_columns = all_ohe_col_names_dict
        self._columns_dict = generated_cols_per_category

        trans_data.loc[:, all_ohe_columns] = np.uint8(0)
        for col in self._categorical_cols:
            features = self._features_dict[col]
            self._set_ohe_column_values(trans_data, col, features, all_ohe_col_names_dict[col])
            del trans_data[col]
        return trans_data

    def get_ohe_col_name_for_categorical_value(self, original_col_name: str, value: str) -> Optional[str]:
        """Returns name of the generated OHE column for given categorical value.
        Returns None if no special OHE column was generated for given value."""
        self._check_transformed()
        if original_col_name not in self._all_ohe_columns:
            raise ValueError(f"No OHE column was generated for column {original_col_name}")
        if pd.isna(value):
            raise ValueError("Value is NA. Use get_ohe_col_name_for_nulls "
                             "if you meant to get column names for NA values.")
        if value in self._all_ohe_columns[original_col_name]:
            return self._all_ohe_columns[original_col_name][value]
        return None

    def get_ohe_col_name_for_nulls(self, original_col_name: str) -> Optional[str]:
        """Returns name of the generated OHE column for NULLs category.
        Returns None if no special OHE column was generated for NULL values."""
        self._check_transformed()
        if original_col_name not in self._all_ohe_columns:
            raise ValueError(f"No OHE column was generated for column {original_col_name}")
        if OneHotEncodingService.NULLS_COL_KEY not in self._all_ohe_columns[original_col_name]:
            return None
        return self._all_ohe_columns[original_col_name][OneHotEncodingService.NULLS_COL_KEY]

    def get_ohe_col_name_for_others(self, original_col_name: str) -> Optional[str]:
        """Returns name of the generated OHE column for Others category.
        Returns None if no special OHE column was generated for extra values."""
        self._check_transformed()
        if original_col_name not in self._all_ohe_columns:
            raise ValueError(f"No OHE column was generated for column {original_col_name}")
        if OneHotEncodingService.OTHERS_COL_KEY not in self._all_ohe_columns[original_col_name]:
            return None
        return self._all_ohe_columns[original_col_name][OneHotEncodingService.OTHERS_COL_KEY]

    def _check_transformed(self):
        if self._all_ohe_columns is None:
            raise ValueError("transform not called yet")

    @staticmethod
    def _get_col_features(data: pd.DataFrame, col: str, threshold: float, max_bin: int) -> List:
        col_params_df = data[col].value_counts(dropna=False)
        col_params_df = col_params_df.reset_index(level=0)
        col_params_df.columns = ["category", "value"]
        col_params_df["percentage"] = col_params_df["value"] / len(data)
        col_params_df["percentage"] = col_params_df["percentage"].cumsum(axis=0)
        number_of_features = OneHotEncodingService._calc_number_of_features(data, col_params_df, threshold, max_bin)
        return col_params_df.iloc[:number_of_features]["category"].to_list()

    @staticmethod
    def _calc_number_of_features(data: pd.DataFrame, col_params: pd.DataFrame, threshold: float, max_bin: int):
        if threshold == 1.0:
            return min(len(data), max_bin)
        if max_bin == INT_MAX:
            return col_params[col_params["percentage"] >= threshold].index[0] + 1
        return min(col_params[col_params["percentage"] >= threshold].index[0] + 1, max_bin)

    @staticmethod
    def _get_all_new_column_names(current_columns_in_frame: List[str], col: str, features: List) -> Dict[str, str]:
        generated_cols = {}
        columns_in_frame = current_columns_in_frame.copy()
        for feature in features:
            if pd.isna(feature):
                col_name = OneHotEncodingService._decide_col_name(columns_in_frame,
                                                                  OneHotEncodingService._get_nulls_col_name(col))
                generated_cols["_$NullValue$_"] = col_name
            else:
                col_name = OneHotEncodingService._decide_col_name(columns_in_frame,
                                                                  OneHotEncodingService._get_feature_col_name(col,
                                                                                                              feature))
                generated_cols[feature] = col_name

            columns_in_frame.append(col_name)

        others_col = OneHotEncodingService._decide_col_name(columns_in_frame,
                                                            OneHotEncodingService._get_others_col_name(col))
        generated_cols["_$OtherValues$_"] = others_col
        return generated_cols

    @staticmethod
    def _decide_col_name(current_columns_in_frame: List[str], col: str) -> str:
        name_index = 2
        col_name = col
        while col_name in current_columns_in_frame:
            col_name = f"{col}{name_index}"
            name_index += 1
        return col_name

    @staticmethod
    def _get_others_col_name(col: str):
        # two underscores used. This makes sure no name collision will occur with other feature columns
        return f"{col}__OTHER"

    @staticmethod
    def _get_nulls_col_name(col: str):
        # two underscores used. This makes sure no name collision will occur with other feature columns
        return f"{col}__NULL"

    @staticmethod
    def _get_feature_col_name(col: str, feature):
        return f"{col}___{feature}"

    @staticmethod
    def _set_ohe_column_values(data: pd.DataFrame, col: str, features: List, ohe_col_names: Dict[str, str]):
        for feature in features:
            if pd.isna(feature):
                transformed_col_name = ohe_col_names[OneHotEncodingService.NULLS_COL_KEY]
                data[transformed_col_name] = np.where(pd.isna(data[col]), 1, 0).astype(np.uint8)
            else:
                transformed_col_name = ohe_col_names[feature]
                data[transformed_col_name] = np.where(data[col] == feature, 1, 0).astype(np.uint8)
        others_col_name = ohe_col_names[OneHotEncodingService.OTHERS_COL_KEY]
        data[others_col_name] = np.where(data[col].isin(features), 0, 1).astype(np.uint8)



"""todo"""
from typing import List

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult


class MainSampleResults:
    """todo"""

    def __init__(self):
        self.comparison_results: List[DqComparisonResult] = None



"""Includes CoarseClassFitOutput class"""
from typing import List, Union

import pandas as pd


class CoarseClassFitOutput:
    """Output of CoarseClass fitting process."""

    def __init__(self):
        self.target_mean: Union[float, int] = None
        self.char_table: pd.DataFrame = None
        self.coarse_class_table: pd.DataFrame = None
        self.rejected_list: List[str] = None



"""Includes UnsupervisedModellingHelper and UnsupervisedFeatureExtractor classes."""
from typing import List

from organon.afe.domain.modelling.unsupervised.afe_unsupervised_algorithm_settings import \
    AfeUnsupervisedAlgorithmSettings
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.modelling.unsupervised_feature_extractor import UnsupervisedFeatureExtractor


class UnsupervisedModellingHelper:
    """
    Helper class for Unsupervised Feature Extraction Service
    """

    @staticmethod
    def get_selected_cols(num_threads: int,
                          algorithm_settings: AfeUnsupervisedAlgorithmSettings = None,
                          frame_with_all_columns: DictDataFrame = None) -> List[str]:
        """Selects best afe columns by unsupervised modelling"""
        input_frame = dict((k, frame_with_all_columns.get_value(k)) for k in frame_with_all_columns.get_column_names())
        extractor = UnsupervisedFeatureExtractor(input_frame, algorithm_settings.bin_count, algorithm_settings.r_factor,
                                                 algorithm_settings.random_state, algorithm_settings.max_column_count,
                                                 algorithm_settings.is_logging)
        return extractor.execute(num_threads)



"""Includes BaseMultipleSourceReportHelperParams class."""
from typing import List

from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.enums.dq_run_type import DqRunType
from organon.idq.domain.reporting.objects.base_dq_report_helper_params import BaseDqReportHelperParams
from organon.idq.domain.settings.partition_info import PartitionInfo


class BaseMultipleCalculationReportHelperParams(BaseDqReportHelperParams):
    """Params for DQ report helper."""

    def __init__(self, data_source_name: str, comparison_results: List[DqComparisonResult],
                 data_column_collection: DqDataColumnCollection, run_type: DqRunType,
                 partition: List[PartitionInfo] = None):
        super().__init__(data_source_name, comparison_results, data_column_collection, run_type, partition=partition)
        self.calculation_count: int = None
        self.comparison_column_info = None
        self.sample_statistics = None
        self.population_statistics = None
        self.data_source_statistics = None



"""Includes DataSourceSettings class."""
from typing import Tuple

from organon.ml.common.enums.classification_type import ClassificationType
from organon.ml.common.enums.color_type import ColorType


class DataSourceSettings:
    """Settings for fine tuning in object classification"""

    def __init__(self, clf_type: ClassificationType, validation_data_ratio: float = 0.2,
                 image_size: Tuple[int, int] = (150, 150),
                 batch_size: int = 50, color_mode: ColorType = ColorType.RGB, random_seed: int = 42):
        self.clf_type = clf_type
        self.validation_data_ratio = validation_data_ratio
        self.image_size = image_size
        self.batch_size = batch_size
        self.color_mode = color_mode
        self.random_seed = random_seed



"""Includes ObjectClassificationConstants"""


class ObjectClassificationConstants:
    """Object classification constants"""
    PREDICTON_COL_NAME = "pred"
    PROBA_COL_NAME = "prob"
    LABELS_COL_NAME = "label"
    FILENAME_COL_NAME = "filename"
    XCEPTION_MIN_IMG_SIZE = 71
    RES_NET_50_V2_MIN_IMG_SIZE = 32
    INCEPTION_RES_NET_V2_MIN_IMG_SIZE = 75



"""Includes DqReportingExecutor class."""
from typing import List, Dict

from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.businessobjects.statistics.data_source_statistics import DataSourceStatistics
from organon.idq.domain.businessobjects.statistics.population_statistics import PopulationStatistics
from organon.idq.domain.businessobjects.statistics.sample_statistics import SampleStatistics
from organon.idq.domain.enums.dq_run_type import DqRunType
from organon.idq.domain.reporting.objects.base_dq_output_report import BaseDqOutputReport
from organon.idq.domain.reporting.objects.single_source_report_helper_params import SingleSourceReportHelperParams
from organon.idq.domain.reporting.objects.source_with_partitions_report_helper_params import \
    SourceWithPartitionsReportHelperParams
from organon.idq.domain.reporting.single_source_report_helper import SingleSourceReportHelper
from organon.idq.domain.reporting.source_with_partitions_report_helper import SourceWithPartitionsReportHelper
from organon.idq.domain.settings.abstractions.dq_full_process_input import DqFullProcessInput


class DqReportingExecutor:
    """Class for execute idq reporting"""

    def __init__(self, full_process_input: DqFullProcessInput, calculation_results: List[DqCalculationResult],
                 comparison_results: List[DqComparisonResult], run_type: DqRunType):
        self.full_process_input = full_process_input
        self.calculation_results = calculation_results
        self.comparison_results = comparison_results
        self.run_type = run_type
        self.t_calc_data_source_name = self.full_process_input.calculation_parameters[
            -1].input_source_settings.source.get_name()
        self.t_calc_col_collection = self.__get_t_calc_data_column_collection()
        self.sample_statistics = self.__get_sample_statistics_dict()
        self.population_statistics = self.__get_population_statistics_dict()
        self.data_source_statistics = self.__get_data_source_statistics_dict()

    def execute(self) -> BaseDqOutputReport:
        """
        Execute the relevant reporting service
        """
        t_calc_partition = self.full_process_input.calculation_parameters[-1].input_source_settings.partition_info_list
        filter_str = self._get_filter_str()
        if self.run_type == DqRunType.RUN_ONE_DATA_SOURCE:
            one_df_dq_report_params = SingleSourceReportHelperParams(self.t_calc_data_source_name,
                                                                     self.comparison_results,
                                                                     self.t_calc_col_collection, self.run_type,
                                                                     partition=t_calc_partition)
            one_df_dq_report_params.filter_str = filter_str
            reporting_helper = SingleSourceReportHelper(one_df_dq_report_params)
        elif self.run_type == DqRunType.RUN_ONE_DATA_SOURCE_WITH_PARTITIONS:
            one_df_with_date_dq_report_params = self._get_source_with_date_report_helper_params()
            reporting_helper = SourceWithPartitionsReportHelper(one_df_with_date_dq_report_params)
        else:
            raise NotImplementedError
        return reporting_helper.execute()

    def _get_source_with_date_report_helper_params(self):
        filter_str = self._get_filter_str()
        t_calc_partition = self.full_process_input.calculation_parameters[-1].input_source_settings.partition_info_list
        one_df_with_date_dq_report_params = SourceWithPartitionsReportHelperParams(self.t_calc_data_source_name,
                                                                                   self.comparison_results,
                                                                                   self.t_calc_col_collection,
                                                                                   self.run_type,
                                                                                   partition=t_calc_partition
                                                                                   )
        one_df_with_date_dq_report_params.calculation_count = len(self.calculation_results)
        one_df_with_date_dq_report_params.comparison_column_info = \
            self.full_process_input.comparison_parameters.comparison_columns
        one_df_with_date_dq_report_params.sample_statistics = self.sample_statistics
        one_df_with_date_dq_report_params.population_statistics = self.population_statistics
        one_df_with_date_dq_report_params.data_source_statistics = self.data_source_statistics
        one_df_with_date_dq_report_params.filter_str = filter_str
        return one_df_with_date_dq_report_params

    def _get_filter_str(self) -> str:
        filter_callable = self.full_process_input.calculation_parameters[-1].input_source_settings.filter_callable
        if filter_callable is not None:
            return filter_callable.__name__
        return None

    def __get_sample_statistics_dict(self) -> Dict[str, SampleStatistics]:
        sample_statistics_dict = {}
        for calculation in self.calculation_results:
            sample_statistics_dict[calculation.calculation_name] = calculation.sample_stats

        return sample_statistics_dict

    def __get_population_statistics_dict(self) -> Dict[str, PopulationStatistics]:
        population_statistics_dict = {}
        for calculation in self.calculation_results:
            population_statistics_dict[calculation.calculation_name] = calculation.population_stats
        return population_statistics_dict

    def __get_data_source_statistics_dict(self) -> Dict[str, DataSourceStatistics]:
        data_source_statistics = {}
        for calculation in self.calculation_results:
            data_source_statistics[calculation.calculation_name] = calculation.data_source_stats

        return data_source_statistics

    def __get_t_calc_data_column_collection(self):
        t_calc_results = self.calculation_results[-1]
        data_column_collection_t = t_calc_results.data_source_stats.data_column_collection
        new_data_col_collection = DqDataColumnCollection()
        for col in data_column_collection_t:
            new_data_col_collection.add(col)

        for calculation_result in self.calculation_results[:-1]:
            for data_column in calculation_result.data_source_stats.data_column_collection:
                if new_data_col_collection.get_column(data_column.column_name) is None:
                    new_data_col_collection.add(data_column)
        return new_data_col_collection



"""Includes TextClassifier class"""
from typing import List, Tuple, Union, Optional

import pandas as pd

from organon.fl.core.helpers import list_helper
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.text_classification.domain.enums.classification_languages import ClassificationLanguages
from organon.ml.text_classification.domain.enums.model_checkpoints import ModelCheckpoints
from organon.ml.text_classification.domain.enums.model_run_type import ModelRunType
from organon.ml.text_classification.domain.enums.selection_metrics import SelectionMetrics
from organon.ml.text_classification.domain.objects.grid_search_settings import GridSearchSettings
from organon.ml.text_classification.domain.objects.model_parameter_settings import ModelParameterSettings
from organon.ml.text_classification.domain.objects.optimizer_settings import OptimizerSettings
from organon.ml.text_classification.domain.objects.text_classification_settings import TextClassificationSettings
from organon.ml.text_classification.domain.objects.tokenizer_settings import TokenizerSettings
from organon.ml.text_classification.domain.services.text_classification_service import TextClassificationService


class TextClassifier:
    """Text Classification UI class"""

    def __init__(self, mdl_run_type: str, checkpoint: str = None, text_column_name: str = None,
                 target_column_name: str = None,
                 language: str = ClassificationLanguages.ENGLISH.name):
        # pylint: disable=too-many-arguments
        model_run_type = get_enum(mdl_run_type.upper(), ModelRunType)
        if checkpoint is not None:
            model_checkpoint = get_enum(checkpoint.upper(), ModelCheckpoints)
        else:
            model_checkpoint = None
        model_language = get_enum(language.upper(), ClassificationLanguages)
        self._settings = TextClassificationSettings(language=model_language,
                                                    text_column=text_column_name,
                                                    target_column=target_column_name,
                                                    mdl_run_type=model_run_type,
                                                    checkpoint=model_checkpoint)
        self._mdl_parameter_settings: ModelParameterSettings = None
        self._optimizer_settings: OptimizerSettings = None
        self._grid_search_settings: GridSearchSettings = None
        self._tokenizer_settings: TokenizerSettings = None
        self._service: TextClassificationService = None
        self.grid_search_params_dict: dict = None
        self.best_model_dict: dict = None

    def fit(self, data: Union[str, Tuple[str, Optional[str]], pd.DataFrame],
            validation_data: Union[str, pd.DataFrame] = None, validation_data_ratio: float = 0.2,
            random_seed: int = None, add_sample_weights: bool = True):
        """fits the classifier"""
        self._set_full_settings()
        self._service = TextClassificationService(self._settings)
        self._service.fit(data, validation_data, validation_data_ratio, random_seed, add_sample_weights)
        self.grid_search_params_dict = self._service.grid_search_params_dict
        self.best_model_dict = self._service.best_params_dict

    def predict(self, test_data: Union[str, Tuple[str, Optional[str]], pd.DataFrame]):
        """prediction of given text example"""
        return self._service.predict(test_data)

    def predict_proba(self, test_data: Union[str, Tuple[str, Optional[str]], pd.DataFrame]):
        """prediction probabilities given text example"""
        return self._service.predict_proba(test_data)

    def set_optimizer_settings(self, learning_rate: float, early_stopping: int, steps_per_epoch: int,
                               early_stopping_min_delta: float):
        """Sets the model optimizer settings for model compilation"""
        self._optimizer_settings = OptimizerSettings(learning_rate, early_stopping, steps_per_epoch,
                                                     early_stopping_min_delta)

    def set_mdl_settings(self, batch_size: int = 20, epoch: int = 20):
        """Sets model parameter settings for model training"""
        self._mdl_parameter_settings = ModelParameterSettings(batch_size, epoch)

    def set_grid_search_settings(self, models: List[str], batch_sizes: List[int],
                                 epochs: List[int],
                                 learning_rates: List[float], early_stopping_patience: List[int],
                                 early_stopping_min_deltas: List[float],
                                 model_selection_metric: str = None):
        # pylint: disable=too-many-arguments
        """Sets grid search parameter settings for High performance classification"""
        model_checkpoints: List[ModelCheckpoints] = []
        if list_helper.is_null_or_empty(models):
            raise ValueError("Checkpoint List can not be empty")

        for checkpoint in models:
            model_checkpoints.append(get_enum(checkpoint.upper(), ModelCheckpoints))

        if model_selection_metric is not None:
            model_selection_metric = get_enum(model_selection_metric.upper(), SelectionMetrics)

        if list_helper.is_null_or_empty(batch_sizes):
            raise ValueError("Batch sizes List can not be empty")
        if list_helper.is_null_or_empty(epochs):
            raise ValueError("Epochs List can not be empty")
        if list_helper.is_null_or_empty(learning_rates):
            raise ValueError("Learning rates List can not be empty")
        if list_helper.is_null_or_empty(early_stopping_patience):
            raise ValueError("Stopping patience List can not be empty")
        if list_helper.is_null_or_empty(early_stopping_min_deltas):
            raise ValueError("Stopping min deltas List can not be empty")
        self._grid_search_settings = GridSearchSettings(model_checkpoints, batch_sizes, epochs, learning_rates,
                                                        early_stopping_patience, early_stopping_min_deltas,
                                                        model_selection_metric)

    def set_tokenizer_settings(self, padding: str, truncation: bool, max_length: int):
        """sets tokenizer settings for text tokenization"""
        self._tokenizer_settings = TokenizerSettings(padding, truncation, max_length)

    def _set_full_settings(self):
        """set all settings specifications of text classification"""
        if self._mdl_parameter_settings is not None:
            self._settings.mdl_param_settings = self._mdl_parameter_settings
        if self._grid_search_settings is not None:
            self._settings.grid_src_settings = self._grid_search_settings
        if self._optimizer_settings is not None:
            self._settings.opt_settings = self._optimizer_settings
        if self._tokenizer_settings is not None:
            self._settings.tokenizer_settings = self._tokenizer_settings



"""
This module keeps the helper functions for list type.
"""
import bisect
from typing import TypeVar, List

T1 = TypeVar('T1')


def binary_search(obj_list: List[T1], value: T1) -> int:
    """
    Returns index of the given value in given sorted list. Returns -1 if not found.
    :param value: value to be searched in list
    :param obj_list: list of objects
    :return: index of given value in list
    """
    search_index = bisect.bisect_left(obj_list, value)
    if search_index != len(obj_list) and obj_list[search_index] == value:
        return search_index
    return -1


def is_null_or_empty(obj_list: List[T1]) -> bool:
    """
    Checks if the object list is null or empty.
    :param obj_list: list of objects
    :return: boolean type to check if the object list is null or empty
    """
    return obj_list is None or len(obj_list) == 0


def is_null(obj: List[T1]) -> bool:
    """
    Checks if the object list is null.
    :param obj: list of objects
    :return: boolean type to check if the object list is null
    """
    return obj is None



"""This module includes Coarse Class input data class."""
from dataclasses import dataclass


@dataclass
class CoarseInputDto:
    """Data class for coarse classing."""
    test_ratio: float
    target_column_type: str
    min_class_size: int
    max_leaf_nodes: int
    stability_check: bool
    stability_threshold: float
    random_state: int
    positive_class: str
    negative_class: str



"""
This module contains AutoColumnTypeDeciderService for detecting quantity and dimension columns.
"""

from typing import Set, List, Tuple, Dict

import pandas as pd

from organon.afe.dataaccess.services.i_sample_data_service import ISampleDataService
from organon.afe.domain.settings.auto_column_decider_settings import AutoColumnDeciderSettings
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.fl.core.helpers import list_helper


class AutoColumnTypeDeciderService:
    """
    This class creates gets sample frame and decides quantity and dimension columns in that frame.
    """

    def __init__(self, sample_service: ISampleDataService,
                 feature_gen_settings: List[FeatureGenerationSettings],
                 trx_entity_col_name: str,
                 auto_decider_settings: AutoColumnDeciderSettings):
        self.frame = None
        self.__sample_service = sample_service

        self.__feature_gen_settings = feature_gen_settings
        self.__trx_entity_column = trx_entity_col_name

        self.__trx_date_columns = [setting.date_column.column_name
                                   for setting in feature_gen_settings
                                   if getattr(setting, "date_column") is not None]

        self.__numeric_to_dimension = auto_decider_settings.numeric_to_dimension
        self.__dimension_distinct_cut_off = auto_decider_settings.dimension_distinct_cut_off

        self.__rejected_quantity_columns = auto_decider_settings.rejected_quantity_columns
        self.__use_quantity_columns = auto_decider_settings.use_quantity_columns

        self.__rejected_dimension_columns = auto_decider_settings.rejected_dimension_columns
        self.__use_dimension_columns = auto_decider_settings.use_dimension_columns

    def decide_quantity_dimension_columns(self) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
        """
        This function handles the logic for auto deciding for quantity and dimension columns
        """
        q_columns_per_setting, d_extra_per_setting = self.decide_quantity_columns()
        d_columns_per_setting = self.decide_dimension_columns(d_extra_per_setting)

        return d_columns_per_setting, q_columns_per_setting

    def decide_quantity_columns(self) -> Tuple[Dict[str, List[str]], Dict[str, Set[str]]]:
        """
        This method handles the logic for deciding quantity columns according to settings
        :return: quantity columns and extra columns to be added to dimension columns
        """
        q_columns_per_setting = {}
        d_extra_per_setting = {}
        q_columns_from_frame = None
        d_extra_from_frame = None

        for order, setting in enumerate(self.__feature_gen_settings):
            q_columns = set()
            d_extra = set()
            if not list_helper.is_null_or_empty(setting.quantity_columns):
                q_columns.update(setting.quantity_columns)
            elif not self.__use_quantity_columns:
                pass
            else:
                if self.frame is None:
                    self.frame = self.__sample_service.get_sample()
                if q_columns_from_frame is None:
                    q_columns_from_frame, d_extra_from_frame = self.__get_quantity_columns()
                q_columns, d_extra = q_columns_from_frame, d_extra_from_frame
            q_columns.difference_update(set(self.__rejected_quantity_columns))
            q_columns_per_setting[order] = list(q_columns)
            d_extra_per_setting[order] = d_extra

        return q_columns_per_setting, d_extra_per_setting

    def decide_dimension_columns(self, d_extra: Dict[str, Set[str]]) -> Dict[str, List[str]]:
        """
        This method handles the logic for deciding dimension columns according to settings
        :param d_extra: extra columns to be added to dimension columns
        :return: list of dimension columns
        """
        d_cols_per_setting = {}
        d_columns_from_frame = None
        for order, setting in enumerate(self.__feature_gen_settings):
            d_columns = set()
            if not list_helper.is_null_or_empty(setting.dimension_columns):
                d_columns.update(setting.dimension_columns)
            elif not self.__use_dimension_columns:
                pass
            else:
                if self.frame is None:
                    self.frame = self.__sample_service.get_sample()
                if d_columns_from_frame is None:
                    d_columns_from_frame = self.__get_dimension_columns()
                d_columns = d_columns_from_frame
            d_columns.update(d_extra[order])
            d_columns.difference_update(set(self.__rejected_dimension_columns))
            d_cols_per_setting[order] = list(d_columns)
        return d_cols_per_setting

    def __get_quantity_columns(self) -> Tuple[Set[str], Set[str]]:
        """
        This function detects quantity columns in frame
        :return: quantity columns and extra columns to be added to dimension columns
        """
        q_columns: Set[str] = set()
        d_extra_columns: Set[str] = set()

        for column_name in self.frame:
            column = self.frame[column_name]
            if not pd.core.dtypes.common.is_datetime_or_timedelta_dtype(column) and \
                    pd.to_numeric(column, errors='coerce').notnull().all():
                distinct_count = column.nunique()
                if distinct_count < self.__numeric_to_dimension and not pd.api.types.is_float(column):
                    d_extra_columns.add(column_name)
                else:
                    q_columns.add(column_name)

        q_columns.difference_update(self.__trx_date_columns + [self.__trx_entity_column])
        d_extra_columns.difference_update(self.__trx_date_columns + [self.__trx_entity_column])

        return q_columns, d_extra_columns

    def __get_dimension_columns(self) -> Set[str]:
        """
        This function detects dimension columns in frame
        :return: dimension columns
        """
        d_columns: Set[str] = set()

        n_rows = len(self.frame)

        for column_name in self.frame:
            column = self.frame[column_name]
            if not pd.api.types.is_numeric_dtype(column) and not pd.api.types.is_datetime64_any_dtype(column):
                d_columns.add(column_name)

        for column_name in d_columns.copy():
            column = self.frame[column_name]
            distinct_count = column.nunique()
            if distinct_count / n_rows > self.__dimension_distinct_cut_off:
                d_columns.remove(column_name)

        d_columns.difference_update(self.__trx_date_columns + [self.__trx_entity_column])

        return d_columns



"""
This module keeps ColumnMeasurementType enum class.
"""
from enum import Enum


class ColumnMeasurementType(Enum):
    """
    Enum class for defining column measurement type values.
    """
    NOT_ASSIGNED = 0
    NOMINAL = 1
    NUMERIC = 2
    OTHER = 3
    DATE = 4
